"""
test gen_data functions
"""
import re

from . import gen_data

NROWS = 58825
ARGS = {
    'nstr': 3,
    'ndate': 2,
    'ndtime': 1,
    'nint': 2,
    'ndouble': 14,
    'nuid': 1
}

def test_df():
    """
    verify if gen_data.df() is generating correct number of rows
    and columns of the expected type
    """
    mydf = gen_data.df(nrows=NROWS, **ARGS)
    nrows = len(mydf)
    assert nrows == NROWS, 'unexpected number of rows %s' % nrows

    cols = mydf.columns
    # verify we have the right number of columns of each type
    for arg in ARGS:
        # chop off leading n
        rx = re.compile(arg[1:], re.I)
        count = 0
        for c in cols:
            if rx.match(c):
                count += 1
        assert count == ARGS[arg], 'unexpected # of %s columns' % arg



# Copyright 2018, Quilt Data Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Present Quilt packages as PyTorch Datasets

`pip install quilt[pytorch]`

Or, in development:
`pip install -e ./[pytorch]`
"""
from torch.utils.data import Dataset

from quilt.nodes import GroupNode

def dataset(
        node_parser,
        include=lambda x: True,
        input_transform=None,
        target_transform=None):
    """Convert immediate children of a GroupNode into a torch.data.Dataset
    Keyword arguments
    * node_parser=callable that converts a DataNode to a Dataset item
    * include=lambda x: True
      lambda(quilt.nodes.GroupNode) => {True, False}
      intended to filter nodes based on metadata
    * input_transform=None; optional callable that takes the item as its argument
    * output_transform=None; optional callable that takes the item as its argument;
      implementation may make its own copy of item to avoid side effects

      Dataset.__getitem__ returns the following tuple
      item = node_parser(node)
      (input_transform(item), output_transform(item))
      Or, if no _transform functions are provided:
      (item, item)
    """
    def _dataset(node, paths): # pylint: disable=unused-argument
        return DatasetFromGroupNode(
            node,
            node_parser=node_parser,
            include=include,
            input_transform=input_transform,
            target_transform=target_transform)

    return _dataset

# pylint: disable=too-few-public-methods
# reason: this interface is baked by torch
class DatasetFromGroupNode(Dataset):
    """Present immediate children of a GroupNode as a torch.dataset"""
    def __init__(
            self,
            group,
            include,
            node_parser,
            input_transform,
            target_transform):

        super(DatasetFromGroupNode, self).__init__()

        if not isinstance(group, GroupNode):
            raise TypeError('Expected group to be GroupNode, got {}', group)
        if not callable(include):
            raise TypeError('Expected include to be callable, got {}', include)

        self.nodes = [x for x in group if include(x)]
        self.node_parser = node_parser
        self.input_transform = input_transform
        self.target_transform = target_transform

    def __getitem__(self, index):
        item = self.node_parser(self.nodes[index])
        target = item
        if self.input_transform:
            item = self.input_transform(item)
        if self.target_transform:
            target = self.target_transform(target)

        return item, target

    def __len__(self):
        return len(self.nodes)



# Copyright 2018, Quilt Data Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Convenience functions for displaying images in Jupyter notebooks.

`pip install quilt[img]`

Or, in development:
`pip install -e ./[img]`
"""

from math import ceil, floor, sqrt
from six import string_types

import matplotlib.image as mpimg
import matplotlib.pyplot as plt

from quilt.tools.const import ELLIPSIS
from quilt.nodes import DataNode, GroupNode
from quilt.tools.build import splitext_no_dot

def plot(figsize=None, formats=None, limit=100, titlelen=10, **kwargs):
    """Display an image [in a Jupyter Notebook] from a Quilt fragment path.
    Intended for use with `%matplotlib inline`.

    Convenience method that loops over supblots that call
    `plt.imshow(image.imread(FRAG_PATH))`.

    Keyword arguments
    * figsize=None # None means auto, else provide (HEIGHT_INCHES, WIDTH_INCHES)
    * formats=None # List of extensions as strings ['jpg', 'png', ...]
    * limit=100 # maximum number of images to display
    * titlelen=10 # max number of characters in subplot title
    * **kwargs - all remaining kwargs are passed to plt.subplots;
      see https://matplotlib.org/api/_as_gen/matplotlib.pyplot.subplots.html
    """
    # pylint: disable=protected-access
    def _plot(node, paths):
        lower_formats = set((x.lower() for x in formats)) if formats is not None else None
        def node_filter(frag, meta):
            filepath = meta.get('_system', {}).get('filepath', None)
            # don't try to read DataFrames as images
            if isinstance(frag, string_types) and filepath:
                _, ext = splitext_no_dot(filepath)
                if lower_formats is None or ext.lower() in lower_formats:
                    return True
            return False
        # assume DataNode has one path; doesn't work with multi-fragment images
        display = [('', paths[0], node._meta)]
        # for GroupNodes, display all DataNode children
        if isinstance(node, GroupNode):
            datanodes = [(x, y) for (x, y) in node._items() if isinstance(y, DataNode)]
            display = [(x, y._data(), y._meta) for (x, y) in datanodes]
            # sort by name so iteration is reproducible (and unit tests pass)
            display = sorted(display, key=lambda rec: rec[0])
            display = [x for x in display if node_filter(x[1], x[2])]
            if len(display) > limit:
                print('Displaying {} of {} images{}'.format(limit, len(display), ELLIPSIS))
                display = display[:limit]
        # display can be empty e.g. if no DataNode children
        if not display:
            print('No images to display.')
            return
        # cast to int to avoid downstream complaints of
        # 'float' object cannot be interpreted as an index
        floatlen = float(len(display)) # prevent integer division in 2.7
        cols = min(int(floor(sqrt(floatlen))), 8)
        rows = int(ceil(floatlen/cols))

        plt.tight_layout()
        plt.subplots(
            rows,
            cols,
            figsize=(cols*2, rows*2) if not figsize else figsize,
            **kwargs)

        for i in range(rows*cols):
            axes = plt.subplot(rows, cols, i + 1) # subplots start at 1, not 0
            axes.axis('off')
            if i < len(display):
                (name, frag, meta) = display[i]
                plt.title(name[:titlelen] + ELLIPSIS if len(name) > titlelen else name)
                filepath = meta.get('_system', {}).get('filepath', None)
                _, ext = splitext_no_dot(filepath)
                try:
                    bits = mpimg.imread(frag, format=ext)
                    plt.imshow(bits)
                # Mac throws OSError, Linux IOError if file not recognizable
                except (IOError, OSError) as err:
                    print('{}: {}'.format(name, str(err)))
                    continue 
    return _plot



# Copyright (c) 2017 Quilt Data, Inc. All rights reserved.

############################################################
# NOTE: This file is shared between compiler and registry. #
# Do not add any client or server specific code here.      #
############################################################

from enum import Enum
import hashlib
import struct

from six import iteritems, itervalues, string_types


LATEST_TAG = 'latest'
README = 'README'


class PackageFormat(Enum):
    HDF5 = 'HDF5'
    PARQUET = 'PARQUET'
    default = PARQUET


class Node(object):
    __slots__ = ('metadata_hash',)

    @property
    @classmethod
    def json_type(cls):
        raise NotImplementedError

    def __init__(self, metadata_hash):
        assert metadata_hash is None or isinstance(metadata_hash, string_types)
        self.metadata_hash = metadata_hash

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__json__() == other.__json__()
        return NotImplemented

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.__json__())

    def __json__(self):
        val = {'type': self.json_type}
        if self.metadata_hash is not None:
            # For backwards compatibility, only store metadata_hash if it's set, to avoid
            # breaking old clients unnecessarily.
            val['metadata_hash'] = self.metadata_hash
        return val

    def get_children(self): # pylint:disable=R0201
        return {}


class GroupNode(Node):
    __slots__ = ('children',)

    json_type = 'GROUP'

    def __init__(self, children, metadata_hash=None):
        super(GroupNode, self).__init__(metadata_hash)
        assert isinstance(children, dict)
        self.children = children

    def get_children(self):
        return self.children

    def __json__(self):
        val = super(GroupNode, self).__json__()
        val['children'] = self.children
        return val


class RootNode(GroupNode):
    __slots__ = ()

    json_type = 'ROOT'


class TableNode(Node):
    __slots__ = ('hashes', 'metadata')

    json_type = 'TABLE'

    def __init__(self, hashes, format, metadata=None, metadata_hash=None):
        super(TableNode, self).__init__(metadata_hash)

        # This AssertionError needed with this message to catch later and present clear user information.
        if not PackageFormat(format) == PackageFormat.PARQUET:
            raise AssertionError("Bad package format '{}', this package may be outdated.".format(format))

        assert isinstance(hashes, list)
        assert isinstance(format, string_types), '%r' % format

        self.hashes = hashes
        self.metadata = metadata or {}

    def __json__(self):
        val = super(TableNode, self).__json__()
        val['hashes'] = self.hashes
        val['metadata'] = self.metadata
        val['format'] = PackageFormat.PARQUET.value
        return val


class FileNode(Node):
    __slots__ = ('hashes', 'metadata')

    json_type = 'FILE'

    def __init__(self, hashes, metadata=None, metadata_hash=None):
        super(FileNode, self).__init__(metadata_hash)

        assert isinstance(hashes, list)

        self.hashes = hashes
        self.metadata = metadata or {}

    def __json__(self):
        val = super(FileNode, self).__json__()
        val['hashes'] = self.hashes
        val['metadata'] = self.metadata
        return val

NODE_TYPE_TO_CLASS = {cls.json_type: cls for cls in [GroupNode, RootNode, TableNode, FileNode]}

def encode_node(node):
    if isinstance(node, Node):
        return node.__json__()
    raise TypeError("Unexpected type: %r" % type(node))

def decode_node(value):
    type_str = value.pop('type', None)
    if type_str is None:
        return value
    node_cls = NODE_TYPE_TO_CLASS[type_str]
    return node_cls(**value)

def hash_contents(contents):
    """
    Creates a hash of key names and hashes in a package dictionary.

    "contents" must be a GroupNode.
    """
    assert isinstance(contents, GroupNode)

    result = hashlib.sha256()

    def _hash_int(value):
        result.update(struct.pack(">L", value))

    def _hash_str(string):
        assert isinstance(string, string_types)
        _hash_int(len(string))
        result.update(string.encode())

    def _hash_object(obj):
        _hash_str(obj.json_type)
        if isinstance(obj, (TableNode, FileNode)):
            hashes = obj.hashes
            _hash_int(len(hashes))
            for hval in hashes:
                _hash_str(hval)
        elif isinstance(obj, GroupNode):
            children = obj.children
            _hash_int(len(children))
            for key, child in sorted(iteritems(children)):
                _hash_str(key)
                _hash_object(child)
        else:
            assert False, "Unexpected object: %r" % obj

        # Backward compatibility: only hash metadata_hash if it's present.
        if obj.metadata_hash is not None:
            _hash_str(obj.metadata_hash)

    _hash_object(contents)

    return result.hexdigest()

def find_object_hashes(root, meta_only=False):
    """
    Iterator that returns hashes of all of the file and table nodes.

    :param root: starting node
    """
    stack = [root]
    while stack:
        obj = stack.pop()
        if not meta_only and isinstance(obj, (TableNode, FileNode)):
            for objhash in obj.hashes:
                yield objhash
        stack.extend(itervalues(obj.get_children()))
        if obj.metadata_hash is not None:
            yield obj.metadata_hash



"""
Integration test of installation from known public package(s)
"""
from .utils import patch

import pytest

from ..tools import command
from .utils import BasicQuiltTestCase
from .integration import skip

def _mock_url(team):
    return pytest.config.getoption("--integration")

@skip
@patch('quilt.tools.command.get_registry_url', _mock_url)
class IntegrationTest(BasicQuiltTestCase):
    def test_env_install(self):
        # public package
        command.install('akarve/days') # package exists on both stage and prod
        from quilt.data.akarve import days
        df = days.names._data()
        # check for expected datum
        assert df.loc[3]['Day'] == 'Wednesday', 'unexpected value in days df at loc[3]'



"""
Unittest setup.
"""

import os
import shutil
from stat import S_IWUSR
import tempfile
import unittest

import quilt
from ..tools.const import PACKAGE_DIR_NAME

try:
    # Python3
    from unittest.mock import patch
except ImportError:
    # Python2 - external dependency.
    from mock import patch

import responses


def quilt_dev_mode(func):
    """Sets quilt._DEV_MODE and restores original state on context exit

    Decorator.

    Use to enable dev mode during a function call -- for example, to
    disable input() prompts during a specific test.
    """
    def decorated(*args, **kwargs):
        dev_mode = quilt._DEV_MODE
        try:
            quilt._DEV_MODE = True
            return func(*args, **kwargs)
        finally:
            quilt._DEV_MODE = dev_mode
    return decorated

def try_require(string):
    """return true iff specified require string resolves properly;
    for use with @pytest.mark.skipif"""
    import pkg_resources
    try:
        pkg_resources.require(string)
    except pkg_resources.ResolutionError as ex:
        print(ex)
        return False
    return True

class BasicQuiltTestCase(unittest.TestCase):
    """
    Base class for unittests.
    - Creates a temporary directory
    """
    def setUp(self):
        self._old_dir = os.getcwd()
        self._test_dir = tempfile.mkdtemp(prefix='quilt-test-')
        os.chdir(self._test_dir)

    def tearDown(self):
        os.chdir(self._old_dir)

        def _onerror(func, path, exc_info):
            """
            Handle read-only files on Windows
            """
            if not os.access(path, os.W_OK):
                os.chmod(path, S_IWUSR)
                func(path)
            else:
                raise
        shutil.rmtree(self._test_dir, onerror=_onerror)

class QuiltTestCase(BasicQuiltTestCase):
    """
    - Mocks requests
    - (And inherits temp directory from superclass)
    """
    def setUp(self):
        super(QuiltTestCase, self).setUp()

        self.auth_patcher = patch('quilt.tools.command._load_auth', lambda: {})
        self.auth_patcher.start()

        self._store_dir = os.path.join(self._test_dir, PACKAGE_DIR_NAME)
        self.store_patcher = patch.dict(os.environ, {'QUILT_PRIMARY_PACKAGE_DIR': self._store_dir})
        self.store_patcher.start()

        self.requests_mock = responses.RequestsMock(assert_all_requests_are_fired=True)
        self.requests_mock.start()

    def tearDown(self):
        self.requests_mock.stop()
        self.auth_patcher.stop()
        self.store_patcher.stop()

        super(QuiltTestCase, self).tearDown()



"""
High-level helper functions for uploading and downloading fragments.
"""

from __future__ import print_function
import gzip
import os
import re
from shutil import copyfileobj, move
import tempfile
from threading import Thread, Lock

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from six import iteritems, itervalues
from tqdm import tqdm

from .hashing import digest_file
from .util import FileWithReadProgress, get_free_space, QuiltException


PARALLEL_UPLOADS = 20
PARALLEL_DOWNLOADS = 20

S3_CONNECT_TIMEOUT = 30
S3_READ_TIMEOUT = 30
S3_TIMEOUT_RETRIES = 3

CONTENT_RANGE_RE = re.compile(r'^bytes (\d+)-(\d+)/(\d+)$')

CHUNK_SIZE = 4096

ZLIB_LEVEL = 2


# pyOpenSSL and S3 don't play well together. pyOpenSSL is completely optional, but gets enabled by requests.
# So... We disable it. That's what boto does.
# https://github.com/boto/botocore/issues/760
# https://github.com/boto/botocore/pull/803
try:
    from urllib3.contrib import pyopenssl
    pyopenssl.extract_from_urllib3()
except ImportError:
    pass


def create_s3_session():
    """
    Creates a session with automatic retries on 5xx errors.
    """
    sess = requests.Session()
    retries = Retry(total=3,
                    backoff_factor=.5,
                    status_forcelist=[500, 502, 503, 504])
    sess.mount('https://', HTTPAdapter(max_retries=retries))
    return sess

def download_fragments(store, obj_urls, obj_sizes):
    assert len(obj_urls) == len(obj_sizes)

    obj_queue = sorted(iteritems(obj_urls), reverse=True)
    total = len(obj_queue)
    # Some objects might be missing a size; ignore those for now.
    total_bytes = sum(size or 0 for size in itervalues(obj_sizes))

    # Check if we have enough disk space. There's no way to check reliably because we also need
    # space for the temporary gzip'ed files, but that's better than nothing.
    free_space = get_free_space(store.object_path('.'))
    if total_bytes > free_space:
        print("Error: Insufficient space for install. Required: %d, available: %d" % (total_bytes, free_space))
        return False

    downloaded = []
    lock = Lock()

    print("Downloading %d fragments (%d bytes before compression)..." % (total, total_bytes))

    with tqdm(total=total_bytes, unit='B', unit_scale=True) as progress:
        def _worker_thread():
            with create_s3_session() as s3_session:
                while True:
                    with lock:
                        if not obj_queue:
                            break
                        obj_hash, url = obj_queue.pop()
                        try:
                            original_size = obj_sizes[obj_hash]
                        except KeyError:
                            raise QuiltException("Malformed package: No size for object " + obj_hash)

                    success = False

                    temp_path_gz = store.temporary_object_path(obj_hash + '.gz')
                    with open(temp_path_gz, 'ab') as output_file:
                        original_last_update = 0
                        for attempt in range(S3_TIMEOUT_RETRIES):
                            try:
                                existing_file_size = output_file.tell()

                                # For zero-byte downloads, we don't need to resume, and range download
                                # must be at least 1 anyways.
                                if original_size == 0:
                                    response = s3_session.get(
                                        url,
                                        timeout=(S3_CONNECT_TIMEOUT, S3_READ_TIMEOUT)
                                    )
                                    if not response.ok:
                                        message = (
                                            "Download failed for {obj_hash}:\n"
                                            "URL: {response.request.url}\n"
                                            "Status code: {response.status_code}\n"
                                            "Response: {response.text!r}\n"
                                        ).format(**locals())  # Splat **kwargs operators ftw
                                        with lock:
                                            tqdm.write(message)
                                        break
                                    if len(response.content):
                                        message = "Expected a zero-byte file, but received content from: "
                                        with lock:
                                            tqdm.write(message + response.url)
                                        break
                                    encoding = response.headers.get('Content-Encoding', None)
                                    # Nothing to write to the filesystem, already created by opening.
                                    success = True
                                    break

                                # Use the Range header to resume downloads.
                                # Weird corner case: if the file is already completely downloaded, we will
                                # get a RANGE_NOT_SATISFIABLE, and not get the Content-Encoding header.
                                # So, always download at least one byte.
                                range_start = max(existing_file_size - 1, 0)

                                response = s3_session.get(
                                    url,
                                    headers={
                                        'Range': 'bytes=%d-' % range_start
                                    },
                                    stream=True,
                                    timeout=(S3_CONNECT_TIMEOUT, S3_READ_TIMEOUT)
                                )

                                if not response.ok:
                                    message = "Download failed for %s:\nURL: %s\nStatus code: %s\nResponse: %r\n" % (
                                        obj_hash, response.request.url, response.status_code, response.text
                                    )
                                    with lock:
                                        tqdm.write(message)
                                    break

                                # Prevent requests from processing 'Content-Encoding: gzip' automatically.
                                encoding = response.raw.headers.pop('Content-Encoding', None)

                                # Make sure we're getting the expected range.
                                content_range = response.headers.get('Content-Range', '')
                                match = CONTENT_RANGE_RE.match(content_range)
                                if not match or not int(match.group(1)) == range_start:
                                    with lock:
                                        tqdm.write("Unexpected Content-Range: %s" % content_range)
                                    break

                                compressed_size = int(match.group(3))

                                # We may have started with a partially-downloaded file, so update the progress bar.
                                compressed_read = existing_file_size
                                original_read = compressed_read * original_size // compressed_size
                                with lock:
                                    progress.update(original_read - original_last_update)
                                original_last_update = original_read

                                # If we've downloaded an extra byte, skip it.
                                skip_bytes = existing_file_size - range_start
                                assert 0 <= skip_bytes <= 1
                                response.raw.read(skip_bytes)

                                # Do the actual download.
                                for chunk in response.iter_content(CHUNK_SIZE):
                                    output_file.write(chunk)
                                    compressed_read += len(chunk)
                                    original_read = compressed_read * original_size // compressed_size
                                    with lock:
                                        progress.update(original_read - original_last_update)
                                    original_last_update = original_read

                                assert output_file.tell() == compressed_read

                                success = True
                                break  # Done!
                            except requests.exceptions.ConnectionError as ex:
                                if attempt < S3_TIMEOUT_RETRIES - 1:
                                    with lock:
                                        tqdm.write("Download for %s timed out; retrying..." % obj_hash)
                                else:
                                    with lock:
                                        tqdm.write("Download failed for %s: %s" % (obj_hash, ex))
                                    break

                    if not success:
                        # We've already printed an error, so not much to do - just move on to the next object.
                        continue

                    if encoding == 'gzip':
                        # Ungzip the downloaded fragment.
                        temp_path = store.temporary_object_path(obj_hash)
                        try:
                            with gzip.open(temp_path_gz, 'rb') as f_in, open(temp_path, 'wb') as f_out:
                                copyfileobj(f_in, f_out)
                        finally:
                            # Delete the file unconditionally - in case it's corrupted and cannot be ungzipped.
                            os.remove(temp_path_gz)
                    elif encoding is None:
                        # The object is not actually compressed.
                        temp_path = temp_path_gz
                    else:
                        tqdm.write("Unexpected encoding for %s: %r" % (obj_hash, encoding))
                        continue

                    # Check the hash of the result.
                    file_hash = digest_file(temp_path)
                    if file_hash != obj_hash:
                        os.remove(temp_path)
                        with lock:
                            tqdm.write("Fragment hashes do not match: expected %s, got %s." %
                                       (obj_hash, file_hash))
                            continue

                    local_filename = store.object_path(obj_hash)
                    move(temp_path, local_filename)

                    # Success.
                    with lock:
                        downloaded.append(obj_hash)

        threads = [
            Thread(target=_worker_thread, name="download-worker-%d" % i)
            for i in range(PARALLEL_DOWNLOADS)
        ]
        for thread in threads:
            thread.daemon = True
            thread.start()
        for thread in threads:
            thread.join()

    return len(downloaded) == total


def upload_fragments(store, obj_urls, obj_sizes, reupload=False):
    assert len(obj_urls) == len(obj_sizes)

    obj_queue = sorted(iteritems(obj_urls), reverse=True)
    total = len(obj_queue)

    total_bytes = sum(itervalues(obj_sizes))

    uploaded = []
    lock = Lock()

    print("Uploading %d fragments (%d bytes)..." % (total, total_bytes))

    with tqdm(total=total_bytes, unit='B', unit_scale=True) as progress:
        def _worker_thread():
            with create_s3_session() as s3_session:
                while True:
                    with lock:
                        if not obj_queue:
                            break
                        obj_hash, obj_urls = obj_queue.pop()
                        original_size = obj_sizes[obj_hash]

                    try:
                        if reupload or not s3_session.head(obj_urls['head']).ok:
                            with FileWithReadProgress(store.object_path(obj_hash), progress.update) as fd:
                                url = obj_urls['put']
                                # Work around a `requests` bug: it treats size 0 as "unknown" and
                                # uses chunked encoding - which S3 doesn't support.
                                data = fd if original_size > 0 else b''
                                response = s3_session.put(url, data=data)
                                response.raise_for_status()
                        else:
                            with lock:
                                tqdm.write("Fragment %s already uploaded; skipping." % obj_hash)
                                progress.update(original_size)

                        with lock:
                            uploaded.append(obj_hash)
                    except requests.exceptions.RequestException as ex:
                        message = "Upload failed for %s:\n" % obj_hash
                        if ex.response is not None:
                            message += "URL: %s\nStatus code: %s\nResponse: %r\n" % (
                                ex.request.url, ex.response.status_code, ex.response.text
                            )
                        else:
                            message += "%s\n" % ex

                        with lock:
                            tqdm.write(message)

        threads = [
            Thread(target=_worker_thread, name="upload-worker-%d" % i)
            for i in range(PARALLEL_UPLOADS)
        ]
        for thread in threads:
            thread.daemon = True
            thread.start()
        for thread in threads:
            thread.join()

    return len(uploaded) == total



"""
Nodes that represent the data in a Quilt package.
"""
import copy
import os

import numpy as np
import pandas as pd
from six import iteritems, itervalues, string_types

from .tools.const import SYSTEM_METADATA, TargetType
from .tools.store import PackageStore
from .tools.util import is_nodename


class Node(object):
    """
    Abstract class that represents a group or a leaf node in a package.
    """
    def __init__(self, meta):
        # Can't instantiate it directly
        assert self.__class__ != Node.__class__
        assert meta is not None
        self._meta = meta

    def _class_repr(self):
        """Only exists to make it easier for subclasses to customize `__repr__`."""
        return "<%s>" % self.__class__.__name__

    def __repr__(self):
        return self._class_repr()

    def __call__(self, asa=None):
        return self._data(asa)

    def _data(self, asa=None):
        raise NotImplementedError


class DataNode(Node):
    """
    Represents a dataframe or a file. Allows accessing the contents using `()`.
    """
    def __init__(self, store, hashes, data, meta):
        super(DataNode, self).__init__(meta)

        self._store = store
        self._hashes = hashes
        self.__cached_data = data

    def __iter__(self):
        # return an empty iterator (not None, not [], which are non-empty)
        return iter(())

    def _target(self):
        return TargetType(self._meta[SYSTEM_METADATA]['target'])

    def _data(self, asa=None):
        """
        Returns the contents of the node: a dataframe or a file path, or passes
        the node and its contents to a callable.
        """
        if asa is not None:
            if self._store is None or self._hashes is None:
                msg = (
                    "Can only use asa functions with built dataframes."
                    " Build this package and try again."
                )
                raise ValueError(msg)
            return asa(self, [self._store.object_path(obj) for obj in self._hashes])
        else:
            if self.__cached_data is None:
                # TODO(dima): Temporary code.
                if self._target() == TargetType.PANDAS:
                    self.__cached_data = self._store.load_dataframe(self._hashes)
                elif self._target() == TargetType.NUMPY:
                    self.__cached_data = self._store.load_numpy(self._hashes)
                else:
                    self.__cached_data = self._store.get_file(self._hashes)
            return self.__cached_data


class GroupNode(Node):
    """
    Represents a group in a package. Allows accessing child objects using the dot notation.
    Warning: calling _data() on a large dataset may exceed local memory capacity in Python (Only
    supported for Parquet packages).
    """
    def __init__(self, meta):
        super(GroupNode, self).__init__(meta)
        self._children = {}

    def __getattr__(self, name):
        if name.startswith('_'):
            raise AttributeError
        else:
            try:
                return self[name]
            except KeyError:
                raise AttributeError

    def __setattr__(self, name, value):
        if name.startswith('_'):
            super(GroupNode, self).__setattr__(name, value)
        else:
            self[name] = value

    def __delattr__(self, name):
        if name.startswith('_'):
            super(GroupNode, self).__detattr__(name, value)
        else:
            try:
                del self[name]
            except KeyError:
                raise AttributeError

    def __dir__(self):
        attrs = set()
        try:
            attrs.update(super(GroupNode, self).__dir__())
        except AttributeError:
            # Fallback for PY2
            attrs.update(dir(type(self)))
            attrs.update(self.__dict__)
        attrs.update(self._children)
        return sorted(attrs)

    def __getitem__(self, name):
        return self._children[name]

    def __setitem__(self, name, value):
        if not isinstance(value, Node):
            raise TypeError("{val} is not a valid package node".format(val=value))
        self._children[name] = value

    def __delitem__(self, name):
        del self._children[name]

    def __contains__(self, name):
        return name in self._children

    def _get(self, name, default=None):
        return self._children.get(name, default)

    def __len__(self):
        return len(self._children)

    def __repr__(self):
        pinfo = super(GroupNode, self).__repr__()
        group_info = '\n'.join(name + '/' for name in sorted(self._group_keys()))
        if group_info:
            group_info += '\n'
        data_info = '\n'.join(sorted(self._data_keys()))
        return '%s\n%s%s' % (pinfo, group_info, data_info)

    def __iter__(self):
        return itervalues(self._children)

    def _items(self):
        return self._children.items()

    def _data_keys(self):
        """
        every child key referencing a dataframe
        """
        return [name for name, child in iteritems(self._children) if not isinstance(child, GroupNode)]

    def _group_keys(self):
        """
        every child key referencing a group that is not a dataframe
        """
        return [name for name, child in iteritems(self._children) if isinstance(child, GroupNode)]

    def _keys(self):
        """
        keys directly accessible on this object via []
        """
        return self._children.keys()

    def _add_group(self, groupname):
        child = GroupNode({})
        self[groupname] = child

    def _data(self, asa=None):
        """
        Merges all child dataframes. Only works for dataframes stored on disk - not in memory.
        """
        hash_list = []
        stack = [self]
        alldfs = True
        store = None
        while stack:
            node = stack.pop()
            if isinstance(node, GroupNode):
                stack.extend(child for _, child in sorted(node._items(), reverse=True))
            else:
                if node._target() != TargetType.PANDAS:
                    alldfs = False
                if node._store is None or node._hashes is None:
                    msg = "Can only merge built dataframes. Build this package and try again."
                    raise NotImplementedError(msg)
                node_store = node._store
                if store is None:
                    store = node_store
                if node_store != store:
                    raise NotImplementedError("Can only merge dataframes from the same store")
                hash_list += node._hashes

        if asa is None:
            if not hash_list:
                return None
            if not alldfs:
                raise ValueError("Group contains non-dataframe nodes")
            return store.load_dataframe(hash_list)
        else:
            if hash_list:
                assert store is not None
                return asa(self, [store.object_path(obj) for obj in hash_list])
            else:
                return asa(self, [])

    def _set(self, path, value, build_dir=''):
        """Create and set a node by path

        This creates a node from a filename or pandas DataFrame.

        If `value` is a filename, it must be relative to `build_dir`.
            `value` is stored as the export path.

        `build_dir` defaults to the current directory, but may be any
            arbitrary directory path, including an absolute path.

        Example:
            # Set `pkg.graph_image` to the data in '/home/user/bin/graph.png'.
            # If exported, it would export to '<export_dir>/bin/graph.png'
            `pkg._set(['graph_image'], 'bin/fizz.bin', '/home/user')`

        :param path:  Path list -- I.e. ['examples', 'new_node']
        :param value:  Pandas dataframe, or a filename relative to build_dir
        :param build_dir:  Directory containing `value` if value is a filename.
        """
        assert isinstance(path, list) and len(path) > 0

        if isinstance(value, pd.DataFrame):
            metadata = {SYSTEM_METADATA: {'target': TargetType.PANDAS.value}}
        elif isinstance(value, np.ndarray):
            metadata = {SYSTEM_METADATA: {'target': TargetType.NUMPY.value}}
        elif isinstance(value, string_types + (bytes,)):
            # bytes -> string for consistency when retrieving metadata
            value = value.decode() if isinstance(value, bytes) else value
            if os.path.isabs(value):
                raise ValueError("Invalid path: expected a relative path, but received {!r}".format(value))
            # Security: filepath does not and should not retain the build_dir's location!
            metadata = {SYSTEM_METADATA: {'filepath': value, 'transform': 'id'}}
            if build_dir:
                value = os.path.join(build_dir, value)
        else:
            accepted_types = tuple(set((pd.DataFrame, np.ndarray, bytes) + string_types))
            raise TypeError("Bad value type: Expected instance of any type {!r}, but received type {!r}"
                            .format(accepted_types, type(value)), repr(value)[0:100])

        for key in path:
            if not is_nodename(key):
                raise ValueError("Invalid name for node: {}".format(key))

        node = self
        for key in path[:-1]:
            child = node._get(key)
            if not isinstance(child, GroupNode):
                child = GroupNode({})
                node[key] = child

            node = child

        key = path[-1]
        node[key] = DataNode(None, None, value, metadata)

    def _filter(self, lambda_or_dict):
        if isinstance(lambda_or_dict, dict):
            func = _create_filter_func(lambda_or_dict)
        elif callable(lambda_or_dict):
            func = lambda_or_dict
        else:
            raise ValueError

        def _filter_node(name, node, func):
            matched = func(node, name)
            if isinstance(node, GroupNode):
                filtered = GroupNode(copy.deepcopy(node._meta))
                for child_name, child_node in node._items():
                    # If the group itself matched, then match all children by using a True filter.
                    child_func = (lambda *args: True) if matched else func
                    filtered_child = _filter_node(child_name, child_node, child_func)
                    if filtered_child is not None:
                        filtered[child_name] = filtered_child

                # Return the group if:
                # 1) It has children, or
                # 2) Group itself matched the filter, or
                # 3) It's the top-level group.
                if matched or len(filtered) or node == self:
                    return filtered
            else:
                if matched:
                    return node
            return None

        return _filter_node('', self, func)


def _create_filter_func(filter_dict):
    filter_name = filter_dict.pop('name', None)
    if filter_name is not None and not isinstance(filter_name, string_types):
        raise ValueError("Invalid 'name': %r" % filter_name)

    filter_meta = filter_dict.pop('meta', None)
    if filter_meta is not None and not isinstance(filter_meta, dict):
        raise ValueError("Invalid 'meta': %r" % filter_meta)

    if filter_dict:
        raise ValueError("Unexpected data in the filter: %r; only 'name' and 'meta' are supported" % filter_dict)

    def helper(value, expected):
        if isinstance(expected, dict):
            if isinstance(value, dict):
                for expected_key, expected_value in iteritems(expected):
                    if not helper(value.get(expected_key), expected_value):
                        return False
                return True
            else:
                return False
        else:
            return value == expected

    def func(node, name):
        if filter_name is not None and filter_name != name:
            return False
        if filter_meta is not None and not helper(node._meta, filter_meta):
            return False
        return True

    return func



"""
Tests for the package store.
"""

import os
import shutil

from ..tools.store import PackageStore, StoreException
from .utils import QuiltTestCase

class StoreTest(QuiltTestCase):
    def test_old_format(self):
        os.mkdir(self._store_dir)
        with open(os.path.join(self._store_dir, '.format'), 'w') as fd:
            fd.write('1.1')

        with self.assertRaises(StoreException):
            PackageStore(self._store_dir)

    def test_teams_migration(self):
        mydir = os.path.dirname(__file__)
        shutil.copytree(os.path.join(mydir, 'store_old_format'), self._store_dir)

        # We're starting with an old version.
        with open(os.path.join(self._store_dir, '.format')) as fd:
            assert fd.read() == '1.2'

        pkg = PackageStore.find_package(None, 'test', 'simple')
        assert pkg is not None

        # We now have a new version.
        with open(os.path.join(self._store_dir, '.format')) as fd:
            assert fd.read() == '1.3'

    def test_buggy_parquet(self):
        mydir = os.path.dirname(__file__)
        shutil.copytree(os.path.join(mydir, 'store_buggy_parquet'), self._store_dir)

        store, pkg = PackageStore.find_package(None, 'test', 'bug')
        obj_hashes = pkg.children['bug'].hashes

        # Make sure this doesn't crash.
        store.load_dataframe(obj_hashes)



import hashlib

from .const import HASH_TYPE

def digest_file(fname):
    """
    Digest files using SHA-2 (256-bit)
    TESTING
      Produces identical output to `openssl sha256 FILE` for the following:
      * on all source .py files and some binary pyc files in parent dir
      * empty files with different names
      * 3.3GB DNAse Hypersensitive file
      * empty file, file with one space, file with one return all produce
      * distinct output
    PERF takes about 20 seconds to hash 3.3GB file
    on an empty file and on build.py
    INSPIRATION: http://stackoverflow.com/questions/3431825/generating-an-md5-checksum-of-a-file
    WARNING: not clear if we need to pad file bytes for proper cryptographic
      hashing
    """
    #chunk size in bytes
    size = 4096
    hval = hashlib.new(HASH_TYPE)
    with open(fname, 'rb') as fd:
        for chunk in iter(lambda: fd.read(size), b''):
            hval.update(chunk)
    return hval.hexdigest()

def digest_string(value):
    hval = hashlib.new(HASH_TYPE)
    hval.update(value.encode('utf8'))
    return hval.hexdigest()



"""test class against quilt.asa.torch"""
import os

import pytest
from six import string_types

from quilt.tools import command
from quilt.nodes import DataNode
from .utils import QuiltTestCase, try_require

if not try_require('quilt[img,pytorch,torchvision]'):
    # pylint: disable=unexpected-keyword-arg
    pytest.skip("only test if [img,pytorch,torchvision] extras installed",
        allow_module_level=True)

# pylint: disable=no-self-use
class ImportTest(QuiltTestCase):
    def test_asa_pytorch(self):
        """test asa.torch interface by converting a GroupNode with asa="""
        from torchvision.transforms import Compose, CenterCrop, ToTensor, Resize
        from torch.utils.data import Dataset
        from PIL import Image
        from torch import Tensor

        from quilt.asa.pytorch import dataset
        # pylint: disable=missing-docstring
        # helper functions to simulate real pytorch dataset usage
        def calculate_valid_crop_size(crop_size, upscale_factor):
            return crop_size - (crop_size % upscale_factor)

        def node_parser(node):
            path = node()
            if isinstance(path, string_types):
                img = Image.open(path).convert('YCbCr')
                chan, _, _ = img.split()
                return chan
            else:
                raise TypeError('Expected string path to an image fragment')

        def input_transform(crop_size, upscale_factor):
            return Compose([
                CenterCrop(crop_size),
                Resize(crop_size // upscale_factor),
                ToTensor(),
            ])

        def target_transform(crop_size):
            def _inner(img):
                img_ = img.copy()
                return Compose([
                    CenterCrop(crop_size),
                    ToTensor(),
                ])(img_)
            return _inner
        # pylint: disable=protected-access
        def is_image(node):
            """file extension introspection on Quilt nodes"""
            if isinstance(node, DataNode):
                filepath = node._meta.get('_system', {}).get('filepath')
                if filepath:
                    return any(
                        filepath.endswith(extension)
                        for extension in [".png", ".jpg", ".jpeg"])
        # end helper functions

        mydir = os.path.dirname(__file__)
        build_path = os.path.join(mydir, 'build_img.yml')
        command.build('foo/torchtest', build_path)
        pkg = command.load('foo/torchtest')

        upscale_factor = 3
        crop_size = calculate_valid_crop_size(256, upscale_factor)
        # pylint: disable=no-member
        my_dataset = pkg.mixed.img(asa=dataset(
            include=is_image,
            node_parser=node_parser,
            input_transform=input_transform(crop_size, upscale_factor),
            target_transform=target_transform(crop_size)
        ))
        assert isinstance(my_dataset, Dataset), \
            'expected type {}, got {}'.format(type(Dataset), type(my_dataset))

        assert my_dataset.__len__() == 2, \
            'expected two images in mixed.img, got {}'.format(my_dataset.__len__())

        for i in range(my_dataset.__len__()):
            tens = my_dataset.__getitem__(i)
            assert all((isinstance(x, Tensor) for x in tens)), \
                'Expected all torch.Tensors in tuple, got {}'.format(tens)



"""
Tests for quilt.tools.util
"""

import os

from .utils import QuiltTestCase
from ..tools.util import sub_dirs, sub_files

class UtilTest(QuiltTestCase):
    def test_sub_files(self):
        mydir = os.path.dirname(__file__)
        path = os.path.join(mydir, './listdir')

        assert set(sub_files(path)) == set(['foo.txt'])
        assert set(sub_files(path, invisible=True)) == set(['.invisible.txt', 'foo.txt'])

    def test_sub_dirs(self):
        mydir = os.path.dirname(__file__)
        path = os.path.join(mydir, './listdir')

        assert set(sub_dirs(path)) == set(['dir'])
        assert set(sub_dirs(path, invisible=True)) == set(['.invisible_dir', 'dir'])



"""
Helper functions.
"""

import numpy as np
from . import nodes
from .tools.command import build_from_node, build, push

def save(package, data, params={}, is_public=False):
    """Build and push data to Quilt registry at user/package/data_node,
    associating params as metadata for the data node.
    :param package: short package specifier string, i.e. 'team:user/pkg/subpath'
    :param data: data to save (np.ndarray or pd.DataFrame)
    :param params: metadata dictionary
    :param is_public: boolean kwarg to push the packages publicly
    """
    for key, value in params.items():
        if isinstance(value, np.ndarray):
            value = value.astype(float)
            params[key] = value.tolist()
    build_from_node(package, nodes.DataNode(None, None, data, params))
    push('{}'.format(package), is_public=is_public)




import re
import pandas as pd
from pandas import DataFrame as df
import numpy

# defined as lowecase globals so importing the library results in clean syntax
# like this:  qc.data[colname], qc.env, qc.filename, etc.
filename = None                 # pylint:disable=C0103
data = None                     # pylint:disable=C0103
env = 'default'                 # pylint:disable=C0103
seed = 0  # for reproducibility             # pylint:disable=C0103

class CheckFunctionsReturn(Exception):
    def __init__(self, result):
        super(CheckFunctionsReturn, self).__init__()
        self.result = result

class CheckFunctionsException(Exception):
    def __init__(self, result):
        super(CheckFunctionsException, self).__init__()
        self.result = result

def check(expr, envs=None):
    global env                  # pylint:disable=C0103
    if envs is not None:
        # TODO: error if env not found
        expr = envs.get(env, expr)
    if not expr:
        raise CheckFunctionsReturn(expr)

def print_recnums(msg, expr, maxrecs=30):
    matching_recs = [str(i) for i, val in enumerate(expr) if val]
    print('{}: {}{}'.format(msg, ",".join(matching_recs[0:maxrecs]),
                            "" if len(matching_recs) < maxrecs else ",..."))

def data_sample(*args, **kwargs):
    global data, seed           # pylint:disable=C0103
    if 'seed' in args:
        seed = kwargs['seed']
    if 'random_state' not in args:
        kwargs['random_state'] = numpy.random.random_sample # pylint:disable=E1101
    data = data.sample(*args, **kwargs)

def check_column_enum(colrx, lambda_or_listexpr, envs=None):
    if envs not in [None, 'default']:
        check_column_regexp(colrx, envs[env])
    for colname in list(data):
        if re.search(colrx, colname):
            if callable(lambda_or_listexpr):
                check(lambda_or_listexpr(data[colname]))
            else:
                check(data[colname].isin(lambda_or_listexpr).all())

VALRANGE_FUNCS = {
    'mean':     lambda col: col.mean(),
    'mode':     lambda col: col.mode(),
    'stddev':   lambda col: col.std(),
    'variance': lambda col: col.var(),
    'median':   lambda col: col.median(),
    'sum':      lambda col: col.sum(),
    'count':    lambda col: col.count(),
}
VALRANGE_FUNCS['avg'] = VALRANGE_FUNCS['mean']
VALRANGE_FUNCS['std'] = VALRANGE_FUNCS['stdev'] = VALRANGE_FUNCS['stddev']
VALRANGE_FUNCS['var'] = VALRANGE_FUNCS['variance']
                
def check_column_valrange(colrx, minval=None, maxval=None, lambda_or_name=None, envs=None):
    if envs not in [None, 'default']:
        check_column_valrange(colrx, minval, maxval, lambda_or_name, envs[env])
    for colname in list(data):
        if re.search(colrx, colname):
            col = data[colname]
            if callable(lambda_or_name):
                col = lambda_or_name(col)
            if minval is None and maxval is None:
                raise CheckFunctionsException(
                    'check_column_valrange() requires minval or maxval')
            minval = col.min() if minval is None else minval
            maxval = col.max() if maxval is None else maxval
            if lambda_or_name in VALRANGE_FUNCS:
                if not check(minval <= VALRANGE_FUNCS[lambda_or_name](col) <= maxval):
                    raise CheckFunctionsReturn(
                        "check_column_valrange column {} out of range {} - {}".format(
                            colname, minval, maxval))
            raise CheckFunctionsException(
                'check_column_valrange(): unknown func: %s' % (lambda_or_name))

def check_column_regexp(colrx, regexp, envs=None):
    if envs not in [None, 'default']:
        check_column_regexp(colrx, regexp, envs[env])
    for colname in list(data):
        if re.search(colrx, colname):
            check(data[colname].astype(str).str.match(regexp).all())

def check_column_substr(colrx, substr, envs=None):
    if envs not in [None, 'default']:
        check_column_substr(colrx, substr, envs[env])
    for colname in list(data):
        if re.search(colrx, colname):
            check( (data[colname].str.index(substr) != -1).all() )

def check_column_datetime(colrx, fmt, envs=None):
    if envs not in [None, 'default']:
        check_column_datetime(colrx, fmt, envs[env])
    for colname in list(data):
        if re.search(colrx, colname):
            try:
                pd.to_datetime(data[colname], format=fmt, errors='raise')
            except Exception as ex:
                raise CheckFunctionsReturn(str(ex))



"""
parse build file, serialize package
"""
from collections import defaultdict, Iterable
import glob
import json
import os
import re

import numpy as np
import pandas as pd
from pandas import DataFrame as df
from six import iteritems, string_types

import yaml
from tqdm import tqdm

from .compat import pathlib
from .const import (DEFAULT_BUILDFILE, PANDAS_PARSERS, DEFAULT_QUILT_YML, PACKAGE_DIR_NAME, RESERVED,
                    SYSTEM_METADATA, QuiltException, TargetType)
from .core import GroupNode, RootNode
from .hashing import digest_file, digest_string
from .store import PackageStore, ParquetLib, StoreException
from .util import (FileWithReadProgress, find_in_package, is_nodename, to_nodename, to_identifier,
                   parse_package)

from . import check_functions as qc            # pylint:disable=W0611


class BuildException(QuiltException):
    """
    Build-time exception class
    """
    pass


def _have_pyspark():
    """
    Check if we're running Pyspark
    """
    if _have_pyspark.flag is None:
        try:
            if PackageStore.get_parquet_lib() is ParquetLib.SPARK:
                import pyspark  # pylint:disable=W0612
                _have_pyspark.flag = True
            else:
                _have_pyspark.flag = False
        except ImportError:
            _have_pyspark.flag = False
    return _have_pyspark.flag
_have_pyspark.flag = None

def _path_hash(path, transform, kwargs):
    """
    Generate a hash of source file path + transform + args
    """
    sortedargs = ["%s:%r:%s" % (key, value, type(value))
                  for key, value in sorted(iteritems(kwargs))]
    srcinfo = "{path}:{transform}:{{{kwargs}}}".format(path=os.path.abspath(path),
                                                   transform=transform,
                                                   kwargs=",".join(sortedargs))
    return digest_string(srcinfo)

def _is_internal_node(node):
    is_leaf = not node or isinstance(node.get(RESERVED['file']), str) or node.get(RESERVED['package'])
    return not is_leaf

def _get_local_args(node, keys):
    result = {}
    for key in keys:
        if key in node:
            # do not consider value as argument in case it has 'file' key
            if (not isinstance(node[key], dict) or
                (isinstance(node[key], dict) and not RESERVED['file'] in node[key])):
                result[key] = node[key]
    return result

def _is_valid_group(group):
    return isinstance(group, dict) or group is None

def _run_checks(dataframe, checks, checks_contents, node_path, rel_path, target, env='default'):
    _ = env  # TODO: env support for checks
    print("Running data integrity checks...")
    checks_list = re.split(r'[,\s]+', checks.strip())
    unknown_checks = set(checks_list) - set(checks_contents)
    if unknown_checks:
        raise BuildException("Unknown check(s) '%s' for %s @ %s" %
                             (", ".join(list(unknown_checks)), rel_path, target.value))
    for check in checks_list:
        res = exec_yaml_python(checks_contents[check], dataframe, node_path, rel_path, target)
        if not res and res is not None:
            raise BuildException("Data check failed: %s on %s @ %s" % (
                check, rel_path, target.value))

def _gen_glob_data(dir, pattern, child_table):
    """Generates node data by globbing a directory for a pattern"""
    dir = pathlib.Path(dir)
    matched = False
    used_names = set()  # Used by to_nodename to prevent duplicate names
    # sorted so that renames (if any) are consistently ordered
    for filepath in sorted(dir.glob(pattern)):
        if filepath.is_dir():
            continue
        else:
            matched = True

        # create node info
        node_table = {} if child_table is None else child_table.copy()
        filepath = filepath.relative_to(dir)
        node_table[RESERVED['file']] = str(filepath)
        node_name = to_nodename(filepath.stem, invalid=used_names)
        used_names.add(node_name)
        print("Matched with {!r}: {!r} from {!r}".format(pattern, node_name, str(filepath)))

        yield node_name, node_table

    if not matched:
        print("Warning: {!r} matched no files.".format(pattern))
        return

def _consume(node, keys):
    for key in keys:
        node.pop(key)

def _build_node(build_dir, pkg_store, pkg_root, node_path, node, checks_contents=None,
                dry_run=False, env='default', ancestor_args={}):
    """
    Parameters
    ----------
    ancestor_args : dict
      any transform inherited from an ancestor
      plus any inherited handler kwargs
      Users can thus define kwargs that affect entire subtrees
      (e.g. transform: csv for 500 .txt files)
      and overriding of ancestor or peer values.
      Child transform or kwargs override ancestor k:v pairs.
    """
    assert isinstance(pkg_store, PackageStore)
    assert isinstance(pkg_root, RootNode)
    if _is_internal_node(node):
        if not dry_run:
            pkg_store.add_to_package_group(pkg_root, node_path, None)

        # Make a consumable copy.  This is to cover a quirk introduced by accepting nodes named
        # like RESERVED keys -- if a RESERVED key is actually matched, it should be removed from
        # the node, or it gets treated like a subnode (or like a node with invalid content)
        node = node.copy()

        # NOTE: YAML parsing does not guarantee key order
        # fetch local transform and kwargs values; we do it using ifs
        # to prevent `key: None` from polluting the update
        local_args = _get_local_args(node, [RESERVED['transform'], RESERVED['kwargs']])
        group_args = ancestor_args.copy()
        group_args.update(local_args)
        _consume(node, local_args)

        # if it's not a reserved word it's a group that we can descend
        groups = {k: v for k, v in iteritems(node) if _is_valid_group(v)}
        _consume(node, groups)

        if node:
            # Unused keys -- either keyword typos or node names with invalid values.
            #   For now, until build.yml schemas, pointing out one should do.
            key, value = node.popitem()
            raise BuildException("Invalid syntax: expected node data for {!r}, got {!r}".format(key, value))
        for child_name, child_table in groups.items():
            if glob.has_magic(child_name):
                # child_name is a glob string, use it to generate multiple child nodes
                for gchild_name, gchild_table in _gen_glob_data(build_dir, child_name, child_table):
                    _build_node(build_dir,
                                pkg_store,
                                pkg_root,
                                node_path + [gchild_name],
                                gchild_table,
                                checks_contents=checks_contents,
                                dry_run=dry_run,
                                env=env,
                                ancestor_args=group_args)
            else:
                if not isinstance(child_name, str) or not is_nodename(child_name):
                    raise StoreException("Invalid node name: %r" % child_name)
                _build_node(build_dir,
                            pkg_store,
                            pkg_root,
                            node_path + [child_name],
                            child_table,
                            checks_contents=checks_contents,
                            dry_run=dry_run,
                            env=env,
                            ancestor_args=group_args)
    else:  # leaf node
        # prevent overwriting existing node names
        if find_in_package(pkg_root, node_path):
            message = "Naming conflict: {!r} added to package more than once"
            raise BuildException(message.format('/'.join(node_path)))
        # handle group leaf nodes (empty groups)
        if not node:
            if not dry_run:
                pkg_store.add_to_package_group(pkg_root, node_path, None)
            return

        include_package = node.get(RESERVED['package'])
        rel_path = node.get(RESERVED['file'])
        if rel_path and include_package:
            message = "A node must define only one of {0} or {1}"
            raise BuildException(message.format(RESERVED['file'], RESERVED['package']))
        elif include_package: # package composition
            team, user, pkgname, subpath = parse_package(include_package, allow_subpath=True)
            store, existing_pkg = PackageStore.find_package(team, user, pkgname)
            if existing_pkg is None:
                raise BuildException("Package not found: %s" % include_package)

            if subpath:
                node = find_in_package(existing_pkg, subpath)
                if node is None:
                    msg = "Package {team}:{owner}/{pkg} has no subpackage: {subpath}"
                    raise BuildException(msg.format(team=team,
                                                    owner=user,
                                                    pkg=pkgname,
                                                    subpath=subpath))
            else:
                node = GroupNode(existing_pkg.children)
            pkg_store.add_to_package_package_tree(pkg_root, node_path, node)
        elif rel_path: # handle nodes built from input files
            path = os.path.join(build_dir, rel_path)

            rel_meta_path = node.get(RESERVED['meta'])
            if rel_meta_path:
                with open(os.path.join(build_dir, rel_meta_path)) as fd:
                    try:
                        metadata = json.load(fd)
                    except ValueError as ex:
                        raise BuildException("Failed to parse %r as JSON: %s" % (rel_meta_path, ex))
                    if SYSTEM_METADATA in metadata:
                        raise BuildException("Invalid metadata in %r: not allowed to use key %r" %
                                             (rel_meta_path, SYSTEM_METADATA))
            else:
                metadata = None

            # get either the locally defined transform and target or inherit from an ancestor
            transform = node.get(RESERVED['transform']) or ancestor_args.get(RESERVED['transform'])

            ID = 'id' # pylint:disable=C0103
            PARQUET = 'parquet' # pylint:disable=C0103
            if transform:
                transform = transform.lower()
                if transform in PANDAS_PARSERS:
                    target = TargetType.PANDAS # pylint:disable=R0204
                elif transform == PARQUET:
                    target = TargetType.PANDAS # pylint:disable=R0204
                elif transform == ID:
                    target = TargetType.FILE # pylint:disable=R0204
                else:
                    raise BuildException("Unknown transform '%s' for %s" %
                                         (transform, rel_path))
            else:
                # Guess transform and target based on file extension if not provided
                _, ext = splitext_no_dot(rel_path)

                if ext in PANDAS_PARSERS:
                    transform = ext
                    target = TargetType.PANDAS
                elif ext == PARQUET:
                    transform = ext
                    target = TargetType.PANDAS
                else:
                    transform = ID
                    target = TargetType.FILE
                print("Inferring 'transform: %s' for %s" % (transform, rel_path))


            # TODO: parse/check environments:
            # environments = node.get(RESERVED['environments'])
            checks = node.get(RESERVED['checks'])
            if transform == ID:
                #TODO move this to a separate function
                if checks:
                    with open(path, 'r') as fd:
                        data = fd.read()
                        _run_checks(data, checks, checks_contents, node_path, rel_path, target, env=env)
                if not dry_run:
                    print("Registering %s..." % path)
                    pkg_store.add_to_package_file(pkg_root,
                                                  path,
                                                  node_path,
                                                  target,
                                                  rel_path,
                                                  transform,
                                                  metadata)
            elif transform == PARQUET:
                if checks:
                    from pyarrow.parquet import ParquetDataset
                    dataset = ParquetDataset(path)
                    table = dataset.read(nthreads=4)
                    dataframe = table.to_pandas()
                    _run_checks(dataframe, checks, checks_contents, node_path, rel_path, target, env=env)
                if not dry_run:
                    print("Registering %s..." % path)
                    pkg_store.add_to_package_file(pkg_root,
                                                  path,
                                                  node_path,
                                                  target,
                                                  rel_path,
                                                  transform,
                                                  metadata)
            else:
                # copy so we don't modify shared ancestor_args
                handler_args = dict(ancestor_args.get(RESERVED['kwargs'], {}))
                # local kwargs win the update
                handler_args.update(node.get(RESERVED['kwargs'], {}))
                # Check Cache
                store = PackageStore()
                path_hash = _path_hash(path, transform, handler_args)
                source_hash = digest_file(path)

                cachedobjs = []
                if os.path.exists(store.cache_path(path_hash)):
                    with open(store.cache_path(path_hash), 'r') as entry:
                        cache_entry = json.load(entry)
                        if cache_entry['source_hash'] == source_hash:
                            cachedobjs = cache_entry['obj_hashes']
                            assert isinstance(cachedobjs, list)

                # TODO: check for changes in checks else use cache
                # below is a heavy-handed fix but it's OK for check builds to be slow
                if (not checks and
                    cachedobjs and
                    all(os.path.exists(store.object_path(obj)) for obj in cachedobjs)):
                    # Use existing objects instead of rebuilding
                    pkg_store.add_to_package_cached_df(pkg_root,
                                                       cachedobjs,
                                                       node_path,
                                                       target,
                                                       rel_path,
                                                       transform,
                                                       metadata)
                else:
                    # read source file into DataFrame
                    print("Serializing %s..." % path)
                    if _have_pyspark():
                        dataframe = _file_to_spark_data_frame(transform, path, handler_args)
                    else:
                        dataframe = _file_to_data_frame(transform, path, handler_args)

                    if checks:
                        # TODO: test that design works for internal nodes... e.g. iterating
                        # over the children and getting/checking the data, err msgs, etc.
                        _run_checks(dataframe, checks, checks_contents, node_path, rel_path, target, env=env)

                    # serialize DataFrame to file(s)
                    if not dry_run:
                        print("Saving as binary dataframe...")
                        obj_hashes = pkg_store.add_to_package_df(pkg_root,
                                                                 dataframe,
                                                                 node_path,
                                                                 target,
                                                                 rel_path,
                                                                 transform,
                                                                 metadata)

                        # Add to cache
                        cache_entry = dict(
                            source_hash=source_hash,
                            obj_hashes=obj_hashes
                            )
                        with open(store.cache_path(path_hash), 'w') as entry:
                            json.dump(cache_entry, entry)
        else: # rel_path and package are both None
            message = "Leaf nodes must define either a %s or %s key"
            raise BuildException(message % (RESERVED['file'], RESERVED['package']))

def _remove_keywords(d):
    """
    copy the dict, filter_keywords

    Parameters
    ----------
    d : dict
    """
    return { k:v for k, v in iteritems(d) if k not in RESERVED }

def _file_to_spark_data_frame(ext, path, handler_args):
    from pyspark import sql as sparksql
    ext = ext.lower() # ensure that case doesn't matter
    logic = PANDAS_PARSERS.get(ext)
    kwargs = dict(logic['kwargs'])
    kwargs.update(handler_args)

    spark = sparksql.SparkSession.builder.getOrCreate()
    dataframe = None
    reader = None
    # FIXME: Add json support?
    if logic['attr'] == "read_csv":
        sep = kwargs.get('sep')
        reader = spark.read.format("csv").option("header", "true")
        if sep:
            reader = reader.option("delimiter", sep)
        dataframe = reader.load(path)

        for col in dataframe.columns:
            pcol = to_identifier(col)
            if col != pcol:
                dataframe = dataframe.withColumnRenamed(col, pcol)
    else:
        dataframe = _file_to_data_frame(ext, path, handler_args)
    return dataframe

def _file_to_data_frame(ext, path, handler_args):
    logic = PANDAS_PARSERS.get(ext)

    # allow user to specify handler kwargs and override default kwargs
    kwargs = logic['kwargs'].copy()
    kwargs.update(handler_args)
    failover = logic.get('failover', None)
    handler = getattr(pd, logic['attr'], None)
    if handler is None:
        raise BuildException("Invalid handler: %r" % logic['attr'])

    dataframe = None
    try_again = False
    try:
        size = os.path.getsize(path)
        with tqdm(total=size, unit='B', unit_scale=True) as progress:
            def _callback(count):
                progress.update(count)
            with FileWithReadProgress(path, _callback) as fd:
                dataframe = handler(fd, **kwargs)
    except ValueError as error:
        if failover:
            warning = "Warning: failed fast parse on input %s.\n" % path
            warning += "Switching to Python engine."
            print(warning)
            try_again = True
        else:
            raise BuildException(str(error))

    if try_again:
        failover_args = {}
        failover_args.update(failover)
        failover_args.update(kwargs)
        try:
            dataframe = handler(path, **failover_args)
        except ValueError as error:
            raise BuildException(str(error))

    # cast object columns to strings
    # TODO does pyarrow finally support objects?
    for name, col in dataframe.iteritems():
        if col.dtype == 'object':
            dataframe[name] = col.astype(str)

    return dataframe

def build_package(team, username, package, subpath, yaml_path,
                  checks_path=None, dry_run=False, env='default'):
    """
    Builds a package from a given Yaml file and installs it locally.

    Returns the name of the package.
    """
    def find(key, value):
        """
        find matching nodes recursively;
        only descend iterables that aren't strings
        """
        if isinstance(value, Iterable) and not isinstance(value, string_types):
            for k, v in iteritems(value):
                if k == key:
                    yield v
                elif isinstance(v, dict):
                    for result in find(key, v):
                        yield result
                elif isinstance(v, list):
                    for item in v:
                        for result in find(key, item):
                            yield result

    build_data = load_yaml(yaml_path)
    # default to 'checks.yml' if build.yml contents: contains checks, but
    # there's no inlined checks: defined by build.yml
    if (checks_path is None and list(find('checks', build_data['contents'])) and
        'checks' not in build_data):
        checks_path = 'checks.yml'
        checks_contents = load_yaml(checks_path, optional=True)
    elif checks_path is not None:
        checks_contents = load_yaml(checks_path)
    else:
        checks_contents = None
    build_package_from_contents(team, username, package, subpath, os.path.dirname(yaml_path), build_data,
                                checks_contents=checks_contents, dry_run=dry_run, env=env)

def get_or_create_package(store, team, owner, pkg, subpath, dry_run=False):
    if subpath:
        root = store.get_package(team, owner, pkg)
        if not root:
            root = store.create_package_node(team, owner, pkg, dry_run=dry_run)
        node = root
        for component in subpath[:-1]:
            node = node.children.setdefault(component, GroupNode({}))
    else:
        root = store.create_package_node(team, owner, pkg, dry_run=dry_run)

    return root

def build_package_from_contents(team, username, package, subpath, build_dir, build_data,
                                checks_contents=None, dry_run=False, env='default'):
    contents = build_data.get('contents', {})
    if not isinstance(contents, dict):
        raise BuildException("'contents' must be a dictionary")

    # inline checks take precedence
    checks_contents = {} if checks_contents is None else checks_contents
    checks_contents.update(build_data.get('checks', {}))

    store = PackageStore()
    newpackage = get_or_create_package(store, team, username, package, subpath, dry_run=dry_run)

    _build_node(build_dir, store, newpackage, subpath, contents,
                checks_contents=checks_contents, dry_run=dry_run, env=env)

    if not dry_run:
        store.save_package_contents(newpackage, team, username, package)

def splitext_no_dot(filename):
    """
    Wrap os.path.splitext to return the name and the extension
    without the '.' (e.g., csv instead of .csv)
    """
    name, ext = os.path.splitext(filename)
    ext = ext.lower()
    return name, ext.strip('.')

def generate_contents(startpath, outfilename=DEFAULT_BUILDFILE):
    """
    Generate a build file (yaml) based on the contents of a
    directory tree.
    """
    def _ignored_name(name):
        return (
            name.startswith('.') or
            name == PACKAGE_DIR_NAME or
            name.endswith('~') or
            name == outfilename
        )

    def _generate_contents(dir_path):
        safename_duplicates = defaultdict(list)
        for name in os.listdir(dir_path):
            if _ignored_name(name):
                continue

            path = os.path.join(dir_path, name)

            if os.path.isdir(path):
                nodename = name
                ext = None
            elif os.path.isfile(path):
                nodename, ext = splitext_no_dot(name)
            else:
                continue

            safename = to_identifier(nodename)
            safename_duplicates[safename].append((name, nodename, ext))

        safename_to_name = {}
        for safename, duplicates in iteritems(safename_duplicates):
            for name, nodename, ext in duplicates:
                if len(duplicates) > 1 and ext:
                    new_safename = to_identifier(name)  # Name with ext
                else:
                    new_safename = safename
                existing_name = safename_to_name.get(new_safename)
                if existing_name is not None:
                    message = "Duplicate node names in directory %r."
                    message += " %r was renamed to %r, which overlaps with %r"
                    raise BuildException(
                        message % (dir_path, name, new_safename, existing_name)
                    )
                safename_to_name[new_safename] = name

        contents = {}
        for safename, name in iteritems(safename_to_name):
            path = os.path.join(dir_path, name)

            if os.path.isdir(path):
                data = _generate_contents(path)
            else:
                rel_path = os.path.relpath(path, startpath)
                data = dict(file=rel_path)

            contents[safename] = data

        return contents

    return dict(
        contents=_generate_contents(startpath)
    )

def generate_build_file(startpath, outfilename=DEFAULT_BUILDFILE):
    """
    Generate a build file (yaml) based on the contents of a
    directory tree.
    """
    buildfilepath = os.path.join(startpath, outfilename)
    if os.path.exists(buildfilepath):
        raise BuildException("Build file %s already exists." % buildfilepath)

    contents = generate_contents(startpath, outfilename)

    with open(buildfilepath, 'w') as outfile:
        yaml.dump(contents, outfile, default_flow_style=False)
    return buildfilepath

def load_yaml(filename, optional=False):
    if optional and (filename is None or not os.path.isfile(filename)):
        return None
    with open(filename, 'r') as fd:
        data = fd.read()
    try:
        res = yaml.safe_load(data)
        if not filename.endswith(DEFAULT_QUILT_YML):
            if 'contents' not in res.keys():
                file_name = os.path.basename(filename)
                raise BuildException('Error in {}: missing "contents" node'.format(file_name))
    except yaml.scanner.ScannerError as error:
        mark = error.problem_mark
        message = ["Bad yaml syntax in {!r}".format(filename),
                   "  Line {}, column {}:".format(mark.line, mark.column)]
        message.extend(error.problem_mark.get_snippet().split(os.linesep))
        message.append("  " + error.problem)
        raise BuildException('\n'.join(message))
    if res is None:
        if optional:
            return None
        raise BuildException("Unable to open YAML file: %s" % filename)
    return res

def exec_yaml_python(chkcode, dataframe, node_path, path, target):
    # TODO False vs Exception...
    try:
        # setup for eval
        qc.nodename = '/'.join(node_path)
        qc.filename = path
        qc.data = dataframe
        eval_globals = {
            'qc': qc, 'numpy': np, 'df': df, 'pd': pd, 're': re
        }
        # single vs multi-line checks - YAML hackery
        if '\n' in str(chkcode):
            # note: python2 doesn't support named args for exec()
            # https://docs.python.org/2/reference/simple_stmts.html#exec
            exec(str(chkcode), eval_globals, {})  # pylint:disable=W0122
            res = True
        else:
            # str() to handle True/False
            res = eval(str(chkcode), eval_globals, {})  # pylint:disable=W0123
    except qc.CheckFunctionsReturn as ex:
        res = ex.result
    except Exception as ex:
        raise BuildException("Data check raised exception: %s on %s @ %s" % (ex, path, target.value))
    return res



"""
generate fake data for testing
"""
import math
import random
import time
import uuid

import numpy as np
import pandas as pd

from ..tools.const import DATEF, TIMEF, DTIMEF

MAXINT = 1000*1000*1000
MINSTR = 5
MAXSTR = 100 # max string length
NCATS = 100 # of types of strings
NROWS = 10000 # 1M
def df(nrows=NROWS,
       nuid=1,
       nstr=1,
       ndate=1,
       ndtime=1,
       nint=2,
       ndouble=7):
    """
    inspiration: https://gist.github.com/wesm/0cb5531b1c2e346a0007
    nuid=a # UUID columns
    strcols=1 # string columns
    datecols=1 # data columns
    dtimecols=0 # date-time columns
    intcols=2 # int columns
    doublecols=7 # double columns

    TODO: use generators over comprehensions for faster perf
    """
    assert nrows > 0, 'rows must be greater than 0'

    cols = []
    data = {}

    for i in range(nuid):
        name = 'UID%s' % nuid
        cols.append(name)
        data[name] = [str(uuid.uuid4()) for _ in range(nrows)]

    for i in range(nstr):
        cats = [
            pd.util.testing.rands(random.randrange(MINSTR, MAXSTR))
            for _ in range(NCATS)
        ]
        # make enough repeats
        ceil = math.ceil(nrows/NCATS)
        # slice to nrows many
        strs = [random.choice(cats) for _ in range(nrows)]
        name = 'Str%s' % i
        cols.append(name)
        data[name] = strs

    for i in range(ndate):
        #get a bunch of unix times
        utimes = [rand_time() for _ in range(nrows)]
        #convert to dates
        maked = lambda t: time.strftime(DATEF, time.gmtime(t))
        dates = [maked(t) for t in utimes]
        name = 'Date%s' % i
        cols.append(name)
        data[name] = dates

    for i in range(ndtime):
        #get a bunch of unix times
        utimes = [rand_time() for _ in range(nrows)]
        #convert to datetimes
        maked = lambda t: time.strftime(DTIMEF, time.gmtime(t))
        dtimes = [maked(t) for t in utimes]
        name = 'DTime%s' % i
        cols.append(name)
        data[name] = dtimes

    for i in range(nint):
        ints = [np.random.randint(-MAXINT, MAXINT) for _ in range(nrows)]
        name = 'Int%s' % i
        cols.append(name)
        data[name] = ints

    for i in range(ndouble):
        doubles = np.random.lognormal(sigma=100, size=nrows)
        name = 'Double%s' % i
        cols.append(name)
        data[name] = pd.Series(doubles)

    return pd.DataFrame(data, columns=cols)


def rand_time():
    """
    random unix time between 0 (epoch start in 1970)
    and today
    """
    maxt = math.floor(time.time())
    return np.random.randint(0, maxt)



"""
Test the build process
"""
import os

import pytest
import numpy as np
import pandas.api.types as ptypes
from pandas.core.frame import DataFrame
from six import assertRaisesRegex, string_types
import yaml

from ..nodes import DataNode, GroupNode
from ..tools.store import ParquetLib, PackageStore
from ..tools.compat import pathlib
from ..tools import build, command, store
from .utils import QuiltTestCase, patch

PACKAGE = 'groot'

class BuildTest(QuiltTestCase):

    def test_build_parquet_default(self):
        """
        Test compilation to Parquet via the default library
        """
        PackageStore.reset_parquet_lib()
        mydir = os.path.dirname(__file__)
        path = os.path.join(mydir, './build_large.yml')
        build.build_package(None, 'test_parquet', PACKAGE, [], path)
        # TODO load DFs based on contents of .yml file at PATH
        # not hardcoded vals (this will require loading modules from variable
        # names, probably using __module__)
        from quilt.data.test_parquet.groot import dataframes, README
        self._test_dataframes(dataframes)
        assert os.path.exists(README())


    def test_build_from_cache(self):
        """
        Build the same package twice and verify that the cache is used and
        that the package is successfully generated.
        """
        mydir = os.path.dirname(__file__)
        path = os.path.join(mydir, './build_large.yml')
        teststore = store.PackageStore()

        # Build once to populate cache
        build.build_package(None, 'test_cache', PACKAGE, [], path)

        # Verify cache contents
        srcpath = os.path.join(mydir, 'data/10KRows13Cols.csv')
        path_hash = build._path_hash(srcpath, 'csv',  {'parse_dates': ['Date0']})
        assert os.path.exists(teststore.cache_path(path_hash))

        # Build again using the cache
        build.build_package(None, 'test_cache', PACKAGE, [], path)

        # TODO load DFs based on contents of .yml file at PATH
        # not hardcoded vals (this will require loading modules from variable
        # names, probably using __module__)
        from quilt.data.test_cache.groot import dataframes, README
        self._test_dataframes(dataframes)
        assert os.path.exists(README())

    def test_parquet_env_var(self):
        """
        Test setting the parquet library using the env variable.
        """
        try:
            assert PackageStore.get_parquet_lib() == ParquetLib.ARROW

            with patch.dict(os.environ, {'QUILT_PARQUET_LIBRARY': ParquetLib.ARROW.value}):
                PackageStore.reset_parquet_lib()
                assert PackageStore.get_parquet_lib() == ParquetLib.ARROW

            with patch.dict(os.environ, {'QUILT_PARQUET_LIBRARY': ParquetLib.SPARK.value}):
                PackageStore.reset_parquet_lib()
                assert PackageStore.get_parquet_lib() == ParquetLib.SPARK
        finally:
            PackageStore.reset_parquet_lib()

    # shared testing logic between pyarrow and default env
    def _test_dataframes(self, dataframes):
        csv = dataframes.csv()
        tsv = dataframes.csv()
        xls = dataframes.xls()
        xls_skip = dataframes.xls_skip()
        rows = len(csv.index)
        assert rows == len(tsv.index) and rows == len(xls.index), \
            'Expected dataframes to have same # rows'
        cols = len(csv.columns)
        assert cols == len(tsv.columns) and cols == len(xls.columns), \
            'Expected dataframes to have same # columns'
        assert xls_skip.shape == (9997, 13), \
            'Expected 9,997 Rows and 13 Columns'
        nulls = dataframes.nulls()
        assert ptypes.is_string_dtype(nulls['strings']), \
            'Expected column of strings to deserialize as strings'
        assert ptypes.is_integer_dtype(nulls['integers']), \
            'Expected column of integers to deserialize as integers'
        assert ptypes.is_float_dtype(nulls['floats']), \
            'Expected column of floats to deserialize as floats'
        assert ptypes.is_numeric_dtype(nulls['integers_nulled']), \
            'Expected column of ints with nulls to deserialize as numeric'
        # TODO add more integrity checks, incl. negative test cases

    def test_build_bad_transform(self):
        path = pathlib.Path(__file__).parent / 'build_bad_transform.yml'

        with pytest.raises(build.BuildException):
            build.build_package(None, 'test_bad_transform', PACKAGE, [], str(path))

    def test_build_bad_file(self):
        # Ensure we generate an error on bad build files
        path = pathlib.Path(__file__).parent / 'build_bad_file.yml'

        with pytest.raises(build.BuildException):
            build.build_package(None, 'test_bad_file', PACKAGE, [], str(path))

    def test_build_empty(self):
        """
        test building from build_empty.yml
        """
        mydir = os.path.dirname(__file__)
        path = os.path.join(mydir, './build_empty.yml')
        build.build_package(None, 'empty', 'pkg', [], path)

        from quilt.data.empty import pkg
        assert not pkg._keys(), 'Expected package to be empty'

    def test_build_reserved(self):
        mydir = os.path.dirname(__file__)
        path = os.path.join(mydir, './build_reserved.yml')
        build.build_package(None, 'reserved', 'pkg', [], path)
        from quilt.data.reserved import pkg
        assert pkg.file, 'Expected package'
        assert pkg.checks, 'Expected package'
        assert pkg.environments, 'Expected package'
        assert pkg.kwargs, 'Expected package'
        assert pkg.transform, 'Expected package'

    def test_build_group_args(self):
        """
        test building from build_group_args.yml
        """
        mydir = os.path.dirname(__file__)
        path = os.path.join(mydir, './build_group_args.yml')
        build.build_package(None, 'groups', 'pkg', [], path)

        from quilt.data.groups import pkg

        assert isinstance(pkg.group_a.csv(), DataFrame), \
            'Expected parent `transform: csv` to affect group_a.csv()'
        assert isinstance(pkg.group_a.tsv(), DataFrame), \
            'Expected local `transform: tsv` to affect group_a.tsv()'
        # TODO these tests should really test the node type and verify it as a file node
        # but currently both raw files and DFs are DataNode instances
        assert isinstance(pkg.group_b.txt(), string_types), \
            'Expected `transform: id` to be inferred from file extension'
        assert isinstance(pkg.group_b.subgroup.txt(), string_types), \
            'Expected `transform: id` to be inferred from file extension'
        # ENDTODO
        assert isinstance(pkg.group_b.tsv(), DataFrame), \
            'Expected `transform: tsv` to be inferred from file extension'
        assert pkg.group_b.tsv()['Date0'].dtype == np.dtype('<M8[ns]'), \
            'Expected Date0 column to parse as date'
        assert pkg.group_b.subgroup.tsv().shape == (1, 3), \
            'Expected `transform: tsv` and one skipped row from group args'
        assert pkg.group_b.subgroup.csv().shape == (0, 2), \
            'Expected local `transform: csv` and one skipped row from group args'
        assert pkg.group_b.subgroup.many_tsv.one().shape == (1, 3), \
            'Expected local `transform: csv` and one skipped row from group args'
        assert isinstance(pkg.group_b.subgroup.many_tsv.two(), DataFrame), \
            'Expected `transform: tsv` from ancestor'
        assert isinstance(pkg.group_b.subgroup.many_tsv.three(), DataFrame), \
            'Expected `transform: tsv` from ancestor'
        assert not pkg.group_empty._keys(), 'Expected group_empty to be empty'
        assert not pkg.group_x.empty_child._keys(), 'Expected group_x.emptychild to be empty'

    def test_generate_buildfile(self):
        """
        Test auto-generating a buildfile for compilation
        """
        mydir = os.path.dirname(__file__)
        path = os.path.join(mydir, 'data')
        buildfilepath = os.path.join(path, 'build.yml')
        assert not os.path.exists(buildfilepath), "%s already exists" % buildfilepath
        build.generate_build_file(path)
        assert os.path.exists(buildfilepath)
        build.build_package(None, 'test_generated', 'generated', [], buildfilepath)
        os.remove(buildfilepath)
        from quilt.data.test_generated.generated import bad, foo, nuts, README # pylint:disable=W0612

    def test_failover(self):
        """
        Test failover to the slower python read-csv on Pandas error
        """
        mydir = os.path.dirname(__file__)
        path = os.path.join(mydir, './build_failover.yml')
        build.build_package(None, 'test_failover', PACKAGE, [], path)
        from quilt.data.test_failover.groot import bad

    def test_duplicates(self):
        mydir = os.path.dirname(__file__)
        path = os.path.join(mydir, 'dups_good')
        buildfilepath = os.path.join(path, 'build.yml')
        if os.path.exists(buildfilepath):
            os.remove(buildfilepath)

        build.generate_build_file(path)
        assert os.path.exists(buildfilepath)

        with open(buildfilepath) as fd:
            docs = yaml.safe_load_all(fd)
            data = next(docs, None)

        contents = data['contents']

        assert contents == {
            # File extensions added to "a" due to conflicts
            'a_txt': {'file': 'a.txt'},
            'a_csv': {'file': 'a.csv'},
            # "a" dir stays the same - but it's fine cause other "a"s got renamed
            'a': {},
            # Directories don't actually have extensions, so include them even with no conficts
            'dir_ext': {},
            # Weird characters replaced with "_"
            'a__b___c': {'file': 'a%%b___c'},
            # Prepend "n" to files that start with a number
            'n1': {'file': '1'},
            # ... even if there used to be an underscore there
            'n123': {'file': '_123'},
            # Handle conflicts with numbers, too
            'n1_txt': {'file': '1.txt'},
        }

        os.remove(buildfilepath)

    def test_duplicates_conflict(self):
        mydir = os.path.dirname(__file__)
        path = os.path.join(mydir, 'dups_bad')
        buildfilename = 'build_test_duplicates_conflict.yml'
        buildfilepath = os.path.join(path, buildfilename)
        if os.path.exists(buildfilepath):
            os.remove(buildfilepath)

        with self.assertRaises(build.BuildException):
            build.generate_build_file(path, outfilename=buildfilename)

    def test_copy(self):
        mydir = os.path.dirname(__file__)
        path = os.path.join(mydir, 'data')
        buildfilename = 'build_test_copy.yml'
        buildfilepath = os.path.join(path, buildfilename)
        assert not os.path.exists(buildfilepath), "%s already exists" % buildfilepath

        command.build_from_path('test_copy/generated', path, outfilename=buildfilename)
        from quilt.data.test_copy.generated import bad, foo, nuts

        assert not os.path.exists(buildfilepath), "%s should not have been created!" % buildfilepath

    def test_build_yaml_syntax_error(self):
        """
        Attempt to build a yml file with a syntax error
        """
        mydir = os.path.dirname(__file__)
        path = os.path.join(mydir, './build_bad_syntax.yml')

        with assertRaisesRegex(self, build.BuildException, r'Bad yaml syntax.*build_bad_syntax\.yml'):
            build.build_package(None, 'test_syntax_error', PACKAGE, [], path)

    def test_build_no_contents_node(self):
        """
        Attempt to build a yml file without contents node
        """
        mydir = os.path.dirname(__file__)
        path = os.path.join(mydir, './build_no_contents_node.yml')

        with assertRaisesRegex(self, build.BuildException, r'Error in build_no_contents_node.yml'):
            build.build_package(None, 'no_contents', PACKAGE, [], path)

    def test_build_checks_yaml_syntax_error(self):    # pylint: disable=C0103
        """
        Attempt to build a yml file with a syntax error
        """
        mydir = os.path.dirname(__file__)
        path = os.path.join(mydir, './build_checks_bad_syntax.yml')
        checks_path = os.path.join(mydir, './checks_bad_syntax.yml')

        with assertRaisesRegex(self, build.BuildException, r'Bad yaml syntax.*checks_bad_syntax\.yml'):
            build.build_package(None, 'test_syntax_error', PACKAGE, [], path, checks_path=checks_path)

    def test_build_glob_naming_conflict(self):
        mydir = pathlib.Path(os.path.dirname(__file__))
        buildfile = mydir / 'build_globbing_name_conflict.yml'

        with pytest.raises(command.CommandException, match="Naming conflict:"):
            command.build('test/globdata', str(buildfile))

    def test_build_via_glob(self):
        # TODO: flesh out this test
        # TODO: remove any unused files from globbing
        mydir = pathlib.Path(os.path.dirname(__file__))
        buildfile = mydir / 'build_globbing.yml'

        command.build('test/globdata', str(buildfile))

        from quilt.data.test import globdata

        # simple checks to ensure files were found and built
        globdata.csv.csv
        globdata.csv.foo
        globdata.csv.nulls
        globdata.csv.nuts
        globdata.csv.n100Rows13Cols
        globdata.csv.subnode.csv
        globdata.csv.subnode.foo
        globdata.csv.subnode.goo
        # excel, kwargs sent
        assert len(globdata.excel.n100Rows13Cols()) == 95
        # naming collision -- acceptable during a single glob specification, should result in a rename
        globdata.collision.csv
        globdata.collision.csv_2

    def test_package_compose(self):
        mydir = pathlib.Path(os.path.dirname(__file__))
        buildfile = mydir / 'build_simple.yml'
        command.build('test/simple', str(buildfile))

        buildfile = mydir / 'build_compose.yml'
        command.build('test/compose1', str(buildfile))

        from quilt.data.test import simple
        from quilt.data.test import compose1

        assert simple.foo().equals(compose1.from_simple_foo())

    def test_compose_package_not_found(self):
        mydir = pathlib.Path(os.path.dirname(__file__))
        buildfile = mydir / 'build_simple.yml'
        command.build('test/simple', str(buildfile))

        missing_dep_build = {
            'contents': {
                'foo': {
                    'package':
                        'test/notapackage'
                    }
                }
            }

        with assertRaisesRegex(self, build.BuildException, r'Package.*not found'):
            build.build_package_from_contents(None, 'test', 'compose2', [], str(mydir), missing_dep_build)

    def test_compose_subpackage_not_found(self):
        mydir = pathlib.Path(os.path.dirname(__file__))
        buildfile = mydir / 'build_simple.yml'
        command.build('test/simple', str(buildfile))

        missing_dep_build = {
            'contents': {
                'foo': {
                    'package':
                        'test/simple/notasubpackage'
                    }
                }
            }

        with assertRaisesRegex(self, build.BuildException, r'Package.*has no subpackage.*'):
            build.build_package_from_contents(None, 'test', 'compose', [], str(mydir), missing_dep_build)

    def test_included_package_is_group_node(self):
        mydir = pathlib.Path(os.path.dirname(__file__))
        buildfile = mydir / 'build_simple.yml'
        command.build('test/simple', str(buildfile))

        build_compose_contents = {
            'contents': {
                'from_simple_foo': {
                    'package': 'test/simple'
                    }
                }
            }
        build.build_package_from_contents(None, 'test', 'compose3', [], str(mydir), build_compose_contents)
        from quilt.data.test import compose3

        assert type(compose3.from_simple_foo) is GroupNode

    def test_group_node_iter(self):
        mydir = pathlib.Path(os.path.dirname(__file__))
        build_compose_contents = {
            'contents': {
                'grp': {
                    'subgrp1': {
                        'data1': {
                            'file': 'data/foo.csv'
                        }
                    },
                    'subgrp2': {
                        'data2': {
                            'file': 'data/foo.csv'
                        }
                    },
                }
            }
        }

        build.build_package_from_contents(None, 'test', 'pkg_node', [], str(mydir), build_compose_contents)

        from quilt.data.test import pkg_node

        for node in pkg_node:
            assert isinstance(node, GroupNode)

        for grps in pkg_node.grp:
            assert isinstance(grps, GroupNode)
            for dat in grps:
                assert isinstance(dat, DataNode)
                for nothing in dat:
                    assert False, 'DataNode should not have iterable children'

    def test_package_and_file_raises_exception(self):
        mydir = pathlib.Path(os.path.dirname(__file__))
        bad_build_contents = {
            'contents': {
                'foo': {
                    'package':
                        'test/simple/notasubpackage',
                    'file':
                        'mydir/myfile.csv'
                    }
                }
            }
        with self.assertRaises(build.BuildException):
            build.build_package_from_contents(None, 'test', 'shouldfail', [], str(mydir), bad_build_contents)

    def test_parquet_source_file(self):
        df = DataFrame(dict(a=[1, 2, 3])) # pylint:disable=C0103
        import pyarrow as pa
        from pyarrow import parquet
        table = pa.Table.from_pandas(df)
        parquet.write_table(table, 'simpledf.parquet')

        build_contents = {
            'contents': {
                'df': {
                    'file': 'simpledf.parquet'
                    }
                }
            }
        build.build_package_from_contents(None, 'test', 'fromparquet', [], '.', build_contents)
        pkg = command.load('test/fromparquet')
        assert df.equals(pkg.df()) # pylint:disable=E1101

    #TODO: Add test for checks on a parquet-sourced dataframe

    def test_subpackage(self):
        mydir = pathlib.Path(os.path.dirname(__file__))

        df = DataFrame(dict(a=[1, 2, 3]))
        arr = np.array([4, 5, 6])
        path = str(mydir / 'build_simple.yml')
        metadata = {'metadata_key': 'metadata_value'}

        command.build('test/foo/a/b/c/empty')
        command.build('test/foo/gn', GroupNode(metadata))
        command.build('test/foo/dn', DataNode(None, None, arr, metadata))
        command.build('test/foo/df', df)
        command.build('test/foo/arr', arr)
        command.build('test/foo/file', path)  # Adds as a plain file
        command.build('test/foo/stuff', path, build_file=True)  # Builds a subpackage

        pkg = command.load('test/foo')
        assert len(pkg.a.b.c.empty) == 0
        assert len(pkg.gn) == 0
        assert all(metadata[k] == pkg.gn._meta[k] for k in metadata)
        assert np.array_equal(pkg.dn(), arr)
        assert all(metadata[k] == pkg.dn._meta[k] for k in metadata)
        assert pkg.df().equals(df)
        assert np.array_equal(pkg.arr(), arr)
        assert pkg.file
        assert pkg.stuff.foo

    def test_invalid_node(self):
        df = DataFrame(dict(a=[1, 2, 3]))
        arr = np.array([4, 5, 6])

        # Cannot build a package out of a data node.
        with self.assertRaises(command.CommandException):
            command.build('test/foo', df)
        with self.assertRaises(command.CommandException):
            command.build('test/foo', arr)

class TestOutput(QuiltTestCase):
    # This is kindof a quirky workaround. 'capsys' and 'capfd' fixtures don't work
    # with unittest.TestCase objects, but they do work with other fixtures -- including
    # those defined in this scope.
    @pytest.fixture(autouse=True)
    def capfd(self, capfd):
        self.capfd = capfd

    def test_read_yaml_exec_flaw(self):
        # We don't execute anything remote, but someone could give a bad build.yml..
        testdir = pathlib.Path(__file__).parent

        with pytest.raises(yaml.constructor.ConstructorError):
            command.build('test/exec_flaw', str(testdir / 'arbitrary_execution.yml'), build_file=True)
        out, err = self.capfd.readouterr()
        assert not "arbitrary code execution" in out
        assert not "arbitrary code execution" in err



"""
Implementation of imports from `quilt.data` and `quilt.team`.

E.g.:
  import quilt.data.$user.$package as $package
  print $package.$table
or
  from quilt.data.$user.$package import $table
  print $table
"""

import imp
import os.path
import sys

from six import iteritems

from .nodes import DataNode, GroupNode
from .tools import core
from .tools.const import SYSTEM_METADATA, TargetType
from .tools.store import PackageStore


class FakeLoader(object):
    """
    Fake module loader used to create intermediate user and package modules.
    """
    def __init__(self, path):
        self._path = path

    def load_module(self, fullname):
        """
        Returns an empty module.
        """
        mod = sys.modules.setdefault(fullname, imp.new_module(fullname))
        mod.__file__ = self._path
        mod.__loader__ = self
        mod.__path__ = []
        mod.__package__ = fullname
        return mod


def _from_core_node(store, core_node):
    if core_node.metadata_hash is not None:
        metadata = store.load_metadata(core_node.metadata_hash)
        assert SYSTEM_METADATA not in metadata
    else:
        metadata = {}

    if isinstance(core_node, (core.TableNode, core.FileNode)):
        metadata[SYSTEM_METADATA] = {
            'filepath': core_node.metadata.get('q_path'),
            'transform': core_node.metadata.get('q_ext'),
            'target':
                TargetType.PANDAS.value
                if isinstance(core_node, core.TableNode)
                else core_node.metadata.get('q_target', TargetType.FILE.value),
        }
        node = DataNode(store, core_node.hashes, None, metadata)
    else:
        if isinstance(core_node, (core.RootNode, core.GroupNode)):
            node = GroupNode(metadata)
        else:
            assert "Unexpected node: %r" % core_node

        for name, core_child in iteritems(core_node.children):
            child = _from_core_node(store, core_child)
            node[name] = child

    return node


class PackageLoader(object):
    """
    Module loader for Quilt tables.
    """
    def __init__(self, store, root):
        self._root = root
        self._store = store

    def load_module(self, fullname):
        """
        Returns an object that lazily looks up tables and groups.
        """
        mod = sys.modules.get(fullname)
        if mod is not None:
            return mod

        # We're creating an object rather than a module. It's a hack, but it's approved by Guido:
        # https://mail.python.org/pipermail/python-ideas/2012-May/014969.html

        mod = _from_core_node(self._store, self._root)
        sys.modules[fullname] = mod
        return mod


class ModuleFinder(object):
    def __init__(self, module, teams):
        self._module_name = module
        self._teams = teams

    """
    Looks up submodules.
    """
    def find_module(self, fullname, path=None):
        """
        Looks up the table based on the module path.
        """
        if not fullname.startswith(self._module_name + '.'):
            # Not a quilt submodule.
            return None

        submodule = fullname[len(self._module_name) + 1:]
        parts = submodule.split('.')

        # Pop the team prefix if this is a team import.
        if self._teams:
            team = parts.pop(0)
        else:
            team = None

        # Handle full paths first.
        if len(parts) == 2:
            store, pkg = PackageStore.find_package(team, parts[0], parts[1])
            if pkg is not None:
                return PackageLoader(store, pkg)
            else:
                return None

        # Return fake loaders for partial paths.
        for store_dir in PackageStore.find_store_dirs():
            store = PackageStore(store_dir)

            if len(parts) == 0:
                assert self._teams
                path = store.team_path(team)
            elif len(parts) == 1:
                path = store.user_path(team, parts[0])

            if os.path.isdir(path):
                return FakeLoader(path)

        # Nothing is found.
        return None



"""
Tests for CLI

Quick reference:
    When adding a new test, the test should append the specific param
        key paths it covers to TESTED_PARAMS.
    When test_cli_new_param or test_cli_missing_param fails, add or
        remove the param key paths from KNOWN_PARAMS, and add/remove
        related tests, which should be easily findable by the key path.
    When test_coverage fails, the path is present in KNOWN_PARAMS, but
        no test has added the key path to TESTED_PARAMS.

=== Adding/Removing Params ===
Terminology:
    param tree: param info represented by an `ArgparseIntrospector`
    keypath: A list, containing indexes into a nested mapping.
        A keypath can be used directly to get an item from an
        `ArgparseIntrospector` object, for example:
            keypath = [0, 'access', 0, 'add', 0]
            params = get_all_param_paths()
            # returns 'package', which is the name of the first param
            # for 'quilt access add <package>, <user>'
            params[keypath]

Usage:
    When a new param is created or deleted in main.py's ArgumentParser,
    a test will fail.  The failed test will show the key paths for which
    a test must be added or removed.

    This indicates that there should be a test written for the specific
    combination of arguments that has been created, and/or that tests
    for arguments that no longer exist should be removed (or verified).

    This helps to recognize when a change has been made that might be a
    breaking change.

    Key paths take the form:
        [0, 'version', 0, 'add', 0]


=== CLI <--> Python API Testing ===

To test that the CLI is calling the API correctly, we mock the
command module using the MockObject class.  During this process,
the following steps occur:
    * params for the specific commands to be tested are marked as
      tested in TESTED_PARAMS.
    * If the OS environment has QUILT_TEST_CLI_SUBPROC=True
      * Quilt is called as a subprocess with the specified options
        and QUILT_TEST_CLI_SUBPROC="True" in the environment
      * QUILT_TEST_CLI_SUBPROC is read and MockObject is used
        by tools/main.py
      * MockObject emits JSON to stdout
      * JSON is read from subprocess
    * Otherwise:
      * quilt.tools.main.command is mocked with MockObject
      * quilt.tools.main.argument_parser() is called
      * quilt.tools.main.main() is called with parsed args
    * Function calls are read from MockObject instance
    * Function is verified to exist in actual command module
    * args and kwargs are bound to the function (without calling it)
      to ensure that the arguments work.
    * Types, values, or other conditions to be sent to the command
      function may be checked by the specific test as well.

The binding step should prevent any CLI parameters from changes or
removal (or incompatible additions) on the API side.
"""
#NOTE: This is slow in subproc mode due to frequent quilt executions (see #194)
#TODO: More tests!
#TODO: remove xfail from test_coverage() once coverage is complete

# BUG: (minor) add args/kwargs items into param trees.
# It's not very easy to do this, but it fixes the issue that
# we currently leave some ambiguity as to whether [0, 'foo', 'bar']
# indicates:
#     Anonymous positional argument 0
#     Followed by required argument 'foo'
#     Followed by required or optional argument 'bar'
# or indicates:
#     Positional 0, when 'foo' is given
#     Followed by required or optional argument "bar"
#
# This means if the argument order is restructured in a particular way,
# a change may not be noticed.  And, as we all know, if it's possible,
# it will eventually occur.
# Fixing this improves our coverage, but might be somewhat complex.

import collections
import inspect
import json
import os
import pkg_resources
import signal
import sys

from subprocess import check_output, CalledProcessError, Popen, PIPE

import pytest
from six import string_types, PY2

from ..tools.const import EXIT_KB_INTERRUPT
from .utils import BasicQuiltTestCase

# inspect.argspec is deprecated, so
try:
    from funcsigs import signature  # python 2.7
except ImportError:
    from inspect import signature


## "Static" vars
_TEST_DIR = os.path.dirname(os.path.abspath(__file__))
_QUILT_DIR = os.path.dirname(_TEST_DIR)
PACKAGE_DIR = os.path.dirname(_QUILT_DIR)

# When a test for CLI params is called, append the param key paths that
# the test addresses to this variable.
# Get an example key path by calling get_all_param_paths()
TESTED_PARAMS = []

## KNOWN_PARAMS
# This is a list of keypaths.
# When adding a new param to the cli, add the param here.
# New or missing cli param keypaths can be found in test errors,
# These can be directly added or removed from the KNOWN_PARAMS
# variable, as befits your situation.
KNOWN_PARAMS = [
    ['--dev'],
    ['--version'],
    [0],
    [0, 'access'],
    [0, 'access', 0],
    [0, 'access', 0, 'add'],
    [0, 'access', 0, 'add', 0],
    [0, 'access', 0, 'add', 1],
    [0, 'access', 0, 'list'],
    [0, 'access', 0, 'list', 0],
    [0, 'access', 0, 'remove'],
    [0, 'access', 0, 'remove', 0],
    [0, 'access', 0, 'remove', 1],
    [0, 'audit'],
    [0, 'audit', 0],
    [0, 'build'],
    [0, 'build', 0],
    [0, 'build', 1],
    [0, 'check'],
    [0, 'check', '--env'],
    [0, 'check', 0],
    [0, 'config'],
    [0, 'delete'],
    [0, 'delete', 0],
    [0, 'export'],
    [0, 'export', 0],
    [0, 'export', 1],
    [0, 'export', '-f'],
    [0, 'export', '-s'],
    [0, 'generate'],
    [0, 'generate', 0],
    [0, 'help'],
    [0, 'help', 0],
    [0, 'inspect'],
    [0, 'inspect', 0],
    [0, 'install'],
    [0, 'install', '-f'],
    [0, 'install', '-m'],
    [0, 'install', '-t'],
    [0, 'install', '-v'],
    [0, 'install', '-x'],
    [0, 'install', 0],
    [0, 'log'],
    [0, 'log', 0],
    [0, 'login'],
    [0, 'login', 0],
    [0, 'logout'],
    [0, 'ls'],
    [0, 'push'],
    [0, 'push', '--public'],
    [0, 'push', '--team'],
    [0, 'push', '--reupload'],
    [0, 'push', '-x'],
    [0, 'push', 0],
    [0, 'rm'],
    [0, 'rm', '-f'],
    [0, 'rm', 0],
    [0, 'search'],
    [0, 'search', 0],
    [0, 'tag'],
    [0, 'tag', 0],
    [0, 'tag', 0, 'add'],
    [0, 'tag', 0, 'add', 0],
    [0, 'tag', 0, 'add', 1],
    [0, 'tag', 0, 'add', 2],
    [0, 'tag', 0, 'list'],
    [0, 'tag', 0, 'list', 0],
    [0, 'tag', 0, 'remove'],
    [0, 'tag', 0, 'remove', 0],
    [0, 'tag', 0, 'remove', 1],
    [0, 'user'],
    [0, 'user', 0],
    [0, 'user', 0, 'create'],
    [0, 'user', 0, 'create', 0],
    [0, 'user', 0, 'create', 1],
    [0, 'user', 0, 'create', 2],
    [0, 'user', 0, 'delete'],
    [0, 'user', 0, 'delete', 0],
    [0, 'user', 0, 'delete', '-f'],
    [0, 'user', 0, 'delete', 1],
    [0, 'user', 0, 'disable'],
    [0, 'user', 0, 'disable', 0],
    [0, 'user', 0, 'disable', 1],
    [0, 'user', 0, 'enable'],
    [0, 'user', 0, 'enable', 0],
    [0, 'user', 0, 'enable', 1],
    [0, 'user', 0, 'list'],
    [0, 'user', 0, 'list', 0],
    [0, 'user', 0, 'reset-password'],
    [0, 'user', 0, 'reset-password', 0],
    [0, 'user', 0, 'reset-password', 1],
    [0, 'version'],
    [0, 'version', 0],
    [0, 'version', 0, 'add'],
    [0, 'version', 0, 'add', 0],
    [0, 'version', 0, 'add', 1],
    [0, 'version', 0, 'add', 2],
    [0, 'version', 0, 'list'],
    [0, 'version', 0, 'list', 0],
]


## Helper functions
def get_all_param_paths():
    """Get all paths by introspecting the ArgumentParser from main()

    This returns a list of key paths into the current param tree
    as per `get_current_param_tree()`.

    Key paths take the form:
        [0, 'version', 0, 'add', 0]
    ..where numbers indicate a positional param, and text indicates either
    a permitted argument for a positional, or an option.

    :returns: list of lists (list of key paths)
    :rtype: list
    """
    return get_missing_key_paths(get_current_param_tree(), {}, exhaustive=True)


def get_current_param_tree():
    """Get a tree of param names.

    This has the same structure as KNOWN_PARAMS, but contains the
    current params from main().

    :rtype: ArgparseIntrospector
    """
    from ..tools import main
    return ArgparseIntrospector(main.argument_parser())


def get_missing_key_paths(a, b, exhaustive=False):
    """Return key paths from a that aren't in b

    Key paths take the form of a list of keys in a which, if followed, lead to
    a missing key in b.

    If 'exhaustive' is True, then paths in `a` whose parents are known to be
    missing from `b` will also be included.  If it is False, only the topmost
    key paths will be included.
    Example:
        if two paths are missing from `b`:  ['x', 'y', 'z'] and ['x', 'y'],
        then with `exhaustive` set, both will be returned.  With `exhaustive`
        not set, only ['x', 'y'] will be returned.

    :param a: ArgparseIntrospector or list of paths to acquire paths from
    :param b: ArgparseIntrospector or list of paths be tested
    :param exhaustive: Include children of paths missing from b as well
    :rtype: list
    """
    if isinstance(a, ArgparseIntrospector):
        a = list(a.iterpaths())
    if isinstance(b, ArgparseIntrospector):
        b = list(b.iterpaths())
    b_set = set(tuple(kp) for kp in b)

    results = []
    for kpath in a:
        if tuple(kpath) not in b_set:
            results.append(kpath)
    return results


def get_paramtest_coverage():
    """Shows CLI param test coverage based on reporting by test functions.

    Returns a dict with `'missing'` and `'percentage'` keys.
    """
    ideal_data = [tuple(v) for v in get_all_param_paths()]
    ideal = set(ideal_data)
    tested = set(tuple(v) for v in TESTED_PARAMS)
    # remove covered test cases that aren't in the main argparser
    tested &= ideal
    missing = ideal - tested
    covered = ideal - missing
    return {'missing': [list(v) for v in ideal_data if v in missing],
            'percentage': len(covered) / len(ideal)}


def fails_binding(func, args, kwargs):
    """Check if the given `args` and `kwargs` fails binding to `func`.

    :returns: False if bound successfully, string message if failed.
    """
    try:
        signature(func).bind(*args, **kwargs)
        return False
    except TypeError as error:
        return error.args[0] if error.args and error.args[0] else "Failed to bind"


class ArgparseIntrospector(collections.Mapping):
    def __init__(self, argparse_object, ignore=('-h', '--help'), leaf_func=lambda x: getattr(x, 'dest', None)):
        """Create a parameter tree from an argparse parser.

        Once an ArgumentParser is configured, it can be passed as the param
        `argparse_object`, which makes the parameters available as a tree.

        If any parameter has an item from `ignore` in its opton strings, it
        will be excluded from the tree and from iteration.

        The `leaf_func` param warrants a bit more of an in-depth description.
        When an object is found not to have any further children, it is passed
        to the leaf_func.  The returned value is then entered into the tree.
        The default function simply returns the object's "dest" attribute,
        which acts as a very short description of what the argument is, in
        this case.

        :param argparse_object: A (configured) ArgumentParser object
        :param ignore: Ignore arguments in given iterable (-h, --help by default)
        :param leaf_func: Receives an object.  Data returned is entered into tree.
        """
        self.data = argparse_object
        self.ignored = ignore
        self.leaf_func = leaf_func

    def __getitem__(self, key):
        """Fetch value by key or key path."""
        if not isinstance(key, list):
            return self._getitem(key)
        key_path = key
        if not key_path:
            raise KeyError("empty key path list")
        value = self
        for k in key_path:
            try:
                value = value[k]
            except (KeyError, TypeError):
                raise KeyError("Invalid key path", key_path)
        return value

    def _getitem(self, k):
        if k in self.ignored:
            raise KeyError("Ignored: {!r}".format(k))

        if isinstance(k, int):
            positionals = self.data._get_positional_actions()
            try:
                action = positionals[k]
            except IndexError:
                raise KeyError(k)
            choices = action.choices
            if not choices:
                return self.leaf_func(action)
            return collections.OrderedDict((choice, ArgparseIntrospector(parser))
                                            for choice, parser in choices.items())
        else:
            optionals = self.data._get_optional_actions()
            for o in optionals:
                if k in o.option_strings:
                    return self.leaf_func(o)
        raise KeyError(k)

    def __len__(self):
        return len(tuple(self.__iter__()))

    def __iter__(self):
        for index, positional in enumerate(self.data._get_positional_actions()):
            yield index
        for optional in self.data._get_optional_actions():
            if any(s in self.ignored for s in optional.option_strings):
                continue
            yield optional.option_strings[0]

    def as_dict(self):
        def _mcopy(old):
            if not old:
                return self.leaf_func(old.data)
            if not isinstance(old, collections.Mapping):
                return old
            new = {}
            for k, v in old.items():
                new[k] = _mcopy(v) if v else None
            return new
        return _mcopy(self)

    @classmethod
    def _iterpaths(cls, mapping, sortkey=lambda x: repr(x)):
        """Iterate recursively over contained keys and key paths"""
        if sortkey:
            keys = sorted(mapping, key=sortkey)
        else:
            keys = mapping.keys()
        for key in keys:
            value = mapping[key]
            yield [key]
            if isinstance(value, collections.Mapping):
                for path in cls._iterpaths(value, sortkey):
                    yield [key] + path

    def iterpaths(self, sortkey=lambda x: repr(x)):
        for x in self._iterpaths(self, sortkey):
            yield x

    def __repr__(self):
        message = "ArgparseIntrospector with args: [\n    {}\n]"
        args = '\n    '.join(repr(x) for x in self.iterpaths())
        return message.format(args)


class MockObject(object):
    """Mocks objects with callables.

    For any attribute that is accessed, if that attribute is callable
    or doesn't exist, it retrieves a fake callabe.
    The fake callable does the following:
      * asserts that self._result is empty
      * records the call args as self._result
      * records whether or not the function existed
      * if use_stdout is True, it prints json of _result to stdout
    """
    _target = None
    __result = {}

    def __init__(self, target, use_stdout=False):
        self._target = target
        self._use_stdout = use_stdout

    @property
    def _result(self):
        """Deletes itself after being read."""
        result = self.__result
        self.__result = {}
        return result

    @_result.setter
    def _result(self, v):
        assert not self.__result
        self.__result = v

    def __getattr__(self, attrname):
        matched = hasattr(self._target, attrname)
        attr = getattr(self._target, attrname, None)

        # don't mock non-callable attributes
        if matched and not callable(attr):
            return attr
        # don't mock exceptions
        if inspect.isclass(attr) and issubclass(attr, BaseException):
            return attr

        def dummy_func(*args, **kwargs):
            bind_failure = fails_binding(attr, args=args, kwargs=kwargs)

            result = {
                'func': attrname,
                'matched': matched,
                'args': args,
                'kwargs': kwargs,
                'bind failure': bind_failure,
            }

            if self._use_stdout:
                print(json.dumps(result, indent=4))
            self._result = result
        return dummy_func


class TestCLI(BasicQuiltTestCase):
    def setUp(self):
        # must be imported from within functions to avoid circular import
        from ..tools import main, command

        self.mock_command = MockObject(command)
        main.command = self.mock_command
        self._main = main

        self.param_tree = get_current_param_tree()
        self.parser = self.param_tree.data

        # used when using subprocess calls
        self.env = os.environ.copy()
        self.env['PYTHON_PATH'] = PACKAGE_DIR

        self.quilt_command = [sys.executable, '-c', 'from quilt.tools import main; main.main()',
                              'quilt testing']
        self.quilt_shell_command = ' '.join([sys.executable, '-c',
                                             '"from quilt.tools import main; main.main()"',
                                             '"quilt testing"'])

    def tearDown(self):
        # restore the real 'command' module back to the 'main' module
        self._main.command = self.mock_command._target

    def execute(self, cli_args):
        """Execute a command using the method specified by the environment

        When "QUILT_TEST_CLI_SUBPROC" is set to "True", use a subprocess.
        Otherwise, call main() directly.

        :returns: dict of return codes and calls made to `command` functions
        """
        # CLI mode -- actually executes "quilt <cli args>"
        # This mode is preferable, once quilt load times improve.
        if self.env.get('QUILT_TEST_CLI_SUBPROC', '').lower() == 'true':
            return self.execute_cli(cli_args)
        # Fast mode -- calls main.main(cli_args) instead of actually executing quilt
        else:
            return self.execute_fast(cli_args)

    def execute_cli(self, cli_args):
        """Execute quilt <cli_args> by executing quilt in a subprocess

        Typically only runs when 'QUILT_TEST_CLI_SUBPROC' is set, and also
        sets it in the subprocess OS environment.

        This method is preferable for completeness of testing, but currently
        quilt loads far too slowly for it to be useful except perhaps in
        automated testing like Travis or Appveyor.
        """
        result = {}

        quilt = self.quilt_command
        cmd = quilt + cli_args
        env = self.env.copy()

        if not env.get('QUILT_TEST_CLI_SUBPROC', '').lower() == 'true':
            env['QUILT_TEST_CLI_SUBPROC'] = "True"

        try:
            result = json.loads(check_output(cmd, env=env).decode())
            result['return code'] = 0
        except CalledProcessError as error:
            result['return code'] = error.returncode
        return result

    def execute_fast(self, cli_args):
        """Execute quilt by calling quilt.tools.main.main(cli_args)

        This process is significantly faster than execute_cli, but may be
        slightly less complete.
        """
        result = {}
        try:
            self._main.main(cli_args)
        except SystemExit as error:
            result['return code'] = error.args[0] if error.args else 0
        else:
            result['return code'] = 0
        result.update(self.mock_command._result)
        return result

    def execute_with_checks(self, cli_args, funcname):
        """Execute via self.execute, then perform basic checks on the results

        Convenience method.

        This may not always be applicable, but it checks a few commond conditions.
        """
        result = self.execute(cli_args)

        assert result['return code'] == 0      # command accepted by argparse?
        assert result['matched'] is True       # found func in mocked object?
        assert not result['bind failure']      # argparse calling args matched func args?
        assert not result['args']              # only kwargs were used to call the function?
        assert result['func'] == funcname      # called func matches expected funcname?

        return result

    def test_cli_new_param(self):
        missing_paths = get_missing_key_paths(self.param_tree, KNOWN_PARAMS, exhaustive=True)
        if missing_paths:
            message = "Unknown/new CLI param key paths:\n\t{}"
            pytest.fail(message.format('\n\t'.join(repr(x) for x in missing_paths)))

    def test_cli_missing_param(self):
        missing_paths = get_missing_key_paths(KNOWN_PARAMS, self.param_tree, exhaustive=True)
        if missing_paths:
            message = "Missing CLI param key paths:\n\t{}"
            pytest.fail(message.format('\n\t'.join(repr(x) for x in missing_paths)))

    def test_cli_command_config(self):
        """Ensures the 'config' command calls a specific API"""
        ## This test covers the following arguments that require testing
        TESTED_PARAMS.extend([
            [0, 'config'],
        ])

        ## This section tests for circumstances expected to be rejected by argparse.
        expect_fail_2_args = [
            'config --badparam'.split(),
            ]
        for args in expect_fail_2_args:
            assert self.execute(args)['return code'] == 2

        ## This section tests for appropriate types and values.
        cmd = ['config']
        result = self.execute_with_checks(cmd, funcname='config')

        assert not result['kwargs']

    def test_cli_command_login(self):
        """Ensures the 'login' command calls a specific API"""
        ## This test covers the following arguments that require testing
        TESTED_PARAMS.extend([
            [0, 'login'],
            [0, 'login', 0],
        ])

        ## This section tests for circumstances expected to be rejected by argparse.
        expect_fail_2_args = [
            'login too many params'.split(),
            ]
        for args in expect_fail_2_args:
            assert self.execute(args)['return code'] == 2, 'with args: ' + str(args)

        ## This section tests for acceptable types and values.
        # plain login
        cmd = ['login']
        result = self.execute_with_checks(cmd, funcname='login')

        assert result['kwargs']['team'] is None

        # login with team name
        cmd = ['login', 'example_team']
        result = self.execute_with_checks(cmd, funcname='login')

        assert result['kwargs'] == {'team': 'example_team'}

    def test_cli_command_logout(self):
        """Ensures the 'login' command calls a specific API"""
        ## This test covers the following arguments that require testing
        TESTED_PARAMS.extend([
            [0, 'logout'],
        ])

        ## This section tests for circumstances expected to be rejected by argparse.
        expect_fail_2_args = [
            'logout too many params'.split(),
            ]
        for args in expect_fail_2_args:
            assert self.execute(args)['return code'] == 2, 'with args: ' + str(args)

        ## This section tests for acceptable types and values.
        cmd = ['logout']
        result = self.execute_with_checks(cmd, funcname='logout')
        assert not result['args']

    def test_cli_command_push(self):
        ## This test covers the following arguments that require testing
        TESTED_PARAMS.extend([
            [0, 'push'],
            [0, 'push', 0],
            [0, 'push', '--public'],
            [0, 'push', '--reupload'],
            [0, 'push', '--team'],
            [0, 'push', '--team', '--public'],
        ])

        ## This section tests for circumstances expected to be rejected by argparse.
        expect_fail_2_args = [
            'push'.split(),
            'push --public'.split(),
            'push --reupload'.split(),
            'push --public --reupload'.split(),
            'push --public --team'.split(),
            ]
        for args in expect_fail_2_args:
            assert self.execute(args)['return code'] == 2, "using args: " + str(args)

        ## This section tests for appropriate types and values.
        cmd = 'push fakeuser/fakepackage'.split()
        result = self.execute_with_checks(cmd, funcname='push')

        assert result['kwargs'] == {
            'reupload': False,
            'is_public': False,
            'package': 'fakeuser/fakepackage',
            'is_team': False,
            'hash' : None,
        }

        ## Test the flags as well..
        # public (and reupload)
        cmd = 'push --reupload --public fakeuser/fakepackage'.split()
        result = self.execute_with_checks(cmd, funcname='push')

        assert result['kwargs'] == {
            'reupload': True,
            'is_public': True,
            'package': 'fakeuser/fakepackage',
            'is_team': False,
            'hash' : None,
        }

        # team (without reupload)
        cmd = 'push --reupload --team blah:fakeuser/fakepackage'.split()
        result = self.execute_with_checks(cmd, funcname='push')

        assert result['kwargs'] == {
            'reupload': True,
            'is_public': False,
            'package': 'blah:fakeuser/fakepackage',
            'is_team': True,
            'hash' : None,
        }

    def test_cli_command_export(self):
        ## This test covers the following arguments that require testing
        TESTED_PARAMS.extend([
            [0, 'export'],
            [0, 'export', 0],
            [0, 'export', 1],
            [0, 'export', '-f'],
            [0, 'export', '-s'],
            ])

        ## This section tests for circumstances expected to be rejected by argparse.
        expect_fail_2_args = [
            'export'.split(),
            'export too many args'.split(),
            ]
        for args in expect_fail_2_args:
            assert self.execute(args)['return code'] == 2

        ## This section tests for appropriate types and values.
        # run the command
        cmd = 'export fakeuser/fakepackage'.split()
        result = self.execute_with_checks(cmd, funcname='export')

        assert result['kwargs'] == {
            'package': 'fakeuser/fakepackage',
            'output_path': '.',
            'force': False,
            'symlinks': False,
        }

        # run command with dest
        cmd = 'export fakeuser/fakepackage fakedir'.split()
        result = self.execute_with_checks(cmd, funcname='export')

        assert result['kwargs'] == {
            'package': 'fakeuser/fakepackage',
            'output_path': 'fakedir',
            'force': False,
            'symlinks': False,
        }

        # run command with force
        cmd = 'export fakeuser/fakepackage fakedir --force'.split()
        result = self.execute_with_checks(cmd, funcname='export')

        assert result['kwargs'] == {
            'package': 'fakeuser/fakepackage',
            'output_path': 'fakedir',
            'force': True,
            'symlinks': False,
        }

        # run command with symlinks
        cmd = 'export fakeuser/fakepackage fakedir --symlinks'.split()
        result = self.execute_with_checks(cmd, funcname='export')

        assert result['kwargs'] == {
            'package': 'fakeuser/fakepackage',
            'output_path': 'fakedir',
            'force': False,
            'symlinks': True,
        }

    def test_cli_option_dev_flag(self):
        # also test ctrl-c
        if os.name == 'nt':
            # Due to how Windows handles ctrl-c events with process groups and consoles,
            # it's not really feasible to test this on Windows because it will want to kill
            # PyTest (and/or the console on the testing system), or to just kill the
            # subprocess (kill -9 equivalent).
            #
            # It *may* be possible if we create a separate terminal for testing, join it,
            # disable ctrl-c events in our own process and our parent process (if any, f.e.
            # when running in appveyor), send a ctrl-c event, then re-enable ctrl-c events
            # for our own and parent process.  ..that *might* work, but I'm not really
            # familiar with the win32 api.
            pytest.xfail("This test is problematic on Windows.")

        TESTED_PARAMS.append(['--dev'])

        SIGINT = signal.SIGINT

        cmd = ['--dev', 'install', 'user/test']
        result = self.execute(cmd)

        # was the --dev arg accepted by argparse?
        assert result['return code'] == 0

        # We need to run a command that blocks.  To do so, I'm disabling the
        # test mocking of the command module, and executing a command that
        # blocks while waiting for input ('config').
        test_environ = os.environ.copy()
        test_environ['QUILT_TEST_CLI_SUBPROC'] = 'false'
        test_environ['PYTHONUNBUFFERED'] = "true"   # prevent blank stdout due to buffering

        # With no '--dev' arg, the process should exit without a traceback
        cmd = self.quilt_command + ['config']
        proc = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE, env=test_environ)

        # Wait for some expected text
        expected = b"Please enter the URL"
        response = proc.stdout.read(len(expected))  # blocks if 'quilt config' produces too little output.
        assert response == expected

        # Send interrupt, and fetch result
        proc.send_signal(SIGINT)
        stdout, stderr = (b.decode() for b in proc.communicate())

        assert 'Traceback' not in stderr
        # Return code should indicate keyboard interrupt
        assert proc.returncode == EXIT_KB_INTERRUPT

        # With the '--dev' arg, the process should display a traceback
        cmd = self.quilt_command + ['--dev', 'config']
        proc = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE, env=test_environ)

        # Wait for some expected text
        expected = b"Please enter the URL"
        response = proc.stdout.read(len(expected))  # blocks if 'quilt config' produces too little output.
        assert response == expected

        # Send interrupt, and check result
        proc.send_signal(SIGINT)
        stdout, stderr = (b.decode() for b in proc.communicate())

        assert 'Traceback (most recent call last)' in stderr
        # Return code should be the generic exit code '1' for unhandled exception
        assert proc.returncode == 1


# need capsys, so this isn't in the unittest class
def test_cli_command_version_flag(capsys):
    """Tests that `quilt --version` is working"""
    TESTED_PARAMS.append(['--version'])

    from quilt.tools.main import main

    with pytest.raises(SystemExit):
        main(['--version'])
    outerr = capsys.readouterr()

    # there's not a lot to test here -- this literally just does the same thing
    # that 'quilt --version' does, but this at least ensures that it still
    # exists and still produces the expected result.
    dist = pkg_resources.get_distribution('quilt')
    expectation = "quilt {} ({})\n".format(dist.version, dist.egg_name())

    # in python 2, apparently argparse's 'version' handler prints to stderr.
    result = outerr.err if PY2 else outerr.out

    assert expectation == result


# need capsys, so this isn't in the unittest class
def test_cli_command_in_help(capsys):
    """Tests for inclusion in 'help'

    Only tests the base subcommand, not sub-subcommands.
    """
    TESTED_PARAMS.append(['--version'])

    from quilt.tools.main import main

    expected_params = set()
    hidden_params = set()
    expected_optionals = set()
    hidden_optionals = {'--dev'}

    for argpath in KNOWN_PARAMS:
        if len(argpath) == 1:
            if isinstance(argpath[0], string_types):
                assert argpath[0].startswith('-'),  "bug in test, not in tested code"
                expected_optionals.add(argpath[0])
            continue
        if isinstance(argpath[1], string_types):
            expected_params.add(argpath[1])

    with pytest.raises(SystemExit):
        main(['help'])

    outerr = capsys.readouterr()
    lines = outerr.out.split('\n')

    for pos, line in enumerate(lines):
        if line.strip() == '<subcommand>':
            start = pos + 1

    found_params = []
    for pos, line in enumerate(lines[start:]):
        print(line)
        if line.strip() == "optional arguments:":
            print("stopping (optionals)")
            start = pos + 1
            break
        splitline = line.split(None, 1)
        if not splitline:
            print('skipped (splitline)')
            continue
        if line[4] == ' ':  # skip multiline help text
            print('skipped (char 4)')
            continue
        arg = splitline[0]
        found_params.append(arg)

    assert found_params == sorted(found_params), 'Help params should be sorted properly'
    assert set(found_params) | hidden_params == expected_params, 'Found params do not match expected params'

    found_optionals = []
    for line in lines[start:]:
        splitline = line.split(None, 1)   # ignore second optional form if present (--help, -h)
        if not splitline:
            continue
        optional = splitline[0].rstrip(',')
        if not optional.startswith('-'):
            continue
        print(optional)
        if optional in ['--help', '-h']:
            continue
        found_optionals.append(optional)

    assert found_optionals == sorted(found_optionals)
    assert set(found_optionals) | hidden_optionals == expected_optionals


@pytest.mark.xfail
def test_coverage():
    result = get_paramtest_coverage()

    # this display can be removed once this function isn't xfailed anymore
    if result['missing']:
        msg = ("\nThe following param paths aren't tested,\n"
               "except for change notification and signature matching:\n  {}")
        print(msg.format("\n  ".join(repr(x) for x in result['missing'])))
    print("Param-path specific test coverage: {:.0f}%".format(result['percentage'] * 100))

    assert not result['missing']  # cli params not tested yet
    assert result['percentage'] == 1



"""
Python 2 / 3 compatibility

In cases where the `six` lib covers an item, use `six`.

For others, import the module here under the python3 name (if there are naming differences).

Unfortunately, this *doesn't* allow for `from .tools.compat.foo import bar`, so if needed,
import frequently-used objects here for convenience.
"""
# unused imports -- pylint: disable=W0611

import sys as _sys


# Reflecting requirements in setup.py
# Python < 3.6
if _sys.version_info < (3, 6):
    import pathlib2 as pathlib
else:
    import pathlib



"""
Test the data checking / linting system
"""
import os
import re

from six import assertRaisesRegex
import yaml
import pytest

from ..tools import build, command
from .utils import QuiltTestCase

def read_yml_file(fn):
    mydir = os.path.dirname(__file__)
    filepath = os.path.join(mydir, fn)
    with open(filepath) as fd:
        return next(yaml.safe_load_all(fd), None)

class ChecksTest(QuiltTestCase):

    def setUp(self):
        super(ChecksTest, self).setUp()
        self.checks_data = read_yml_file('checks_simple.yml')
        self.checks_contents = self.checks_data['contents']
        self.build_data = read_yml_file('build_simple_checks.yml')
        self.build_contents = self.build_data['contents']

    def build_success(self, check, nodename='foo'):
        if check is not None:
            self.build_contents[nodename]['checks'] = check
        if check == 'string':
            raise Exception(str(self.build_data)+"\n\n"+str(self.checks_contents))
        mydir = os.path.dirname(__file__)
        build.build_package_from_contents(
            None, 'foox', 'barx', [], mydir, self.build_data, self.checks_contents, dry_run=True)

    def build_fail(self, check, regexp=None, nodename='foo'):
        # explicitly fail if build_success() somehow succeeds when it shouldn't
        try:
            self.build_success(check, nodename=nodename)
            pytest.fail('build_fail() called but test succeeded.')
        except pytest.fail.Exception:
            raise
        except Exception:
            pass
        if regexp is None:
            regexp = "Data check failed: %s" % (check)
        with assertRaisesRegex(self, build.BuildException, regexp):
            self.build_success(check, nodename=nodename)

    def test_parse_checks_file(self):
        assert str(self.checks_contents['negative']) == 'False'
        assert self.checks_contents['simple_multiline'] == '# comment\nqc.check(True)\n'
        assert self.build_contents['foo']['checks'] == 'simple, inline_only, inline_and_external'

    def test_bad_checks_reference(self):
        self.build_fail('doesnt_exist', "Unknown check.+doesnt_exist")

    def test_external_only(self):
        del self.build_data['checks']
        self.build_fail('inline_only', "Unknown check.+inline_only")

    def test_inline_only(self):
        self.checks_contents = self.checks_data = None
        self.build_success('inline_only')
        self.build_fail('hasrecs', "Unknown check.+hasrecs")

    def test_simple_checks(self):
        self.build_success('simple')
        self.build_success('simple_multiline')
        self.build_fail('negative')
        self.build_success('hasrecs')
        self.build_success('hascols')
        self.build_success('cardinality')
        self.build_fail('has9999cols')
        self.build_success('multiline_success')
        self.build_fail('multiline_fail')
        self.build_success('enum_list_success')
        self.build_fail('enum_list_fail')
        self.build_fail('enum_list_fail_empty')
        self.build_success('enum_lambda_success')
        self.build_success('stddev')
        self.build_success('sum')

    def test_inline_vs_external(self):
        self.build_success('inline_only')
        self.build_success('inline_and_external')
        self.build_success('hasrecs')

    def test_many_errors(self):
        # TODO: capture details by line number
        self.build_contents['foo'] = { 'checks': 'lots_uid_errors',
                                       'file': 'data/10KRows13Cols.csv' }
        self.checks_contents['lots_uid_errors'] = (
            "qc.check_column_regexp('UID1', r'^[0-9a-e]')")
        self.build_fail('lots_uid_errors')

    def test_string_checks(self):
        self.build_contents['foo'] = { 'checks': 'stringchecks',
                                       'file': 'data/10KRows13Cols.tsv' }
        self.checks_contents['stringchecks'] = """
qc.check_column_substr('UID.+', '-')  # must contain a dash
qc.check_column_regexp('UID.+', r'^[0-9a-f-]+$')
qc.check_column_datetime('Date0', '%Y-%m-%d')
qc.check_column_datetime('DTime0', '%Y-%m-%d %H:%M:%S.%f')
        """
        self.build_success('stringchecks')

    def test_empty_checks(self):
        self.build_contents = {}
        self.checks_contents = None
        self.build_success(None)

    def test_build_package(self):
        def run_build(build_fn=None, checks_fn=None, expect_error=False):
            build_fn = os.path.join(os.path.dirname(__file__), build_fn) if build_fn else None
            checks_fn = os.path.join(os.path.dirname(__file__), checks_fn) if checks_fn else None
            if expect_error:
                with assertRaisesRegex(self, IOError, 'doesnt_exist.yml'):
                    build.build_package(None, 'foox', 'barx', [], build_fn, checks_fn, dry_run=True)
            else:
                build.build_package(None, 'foox', 'barx', [], build_fn, checks_fn, dry_run=True)
        run_build("build_simple_checks.yml", "checks_simple.yml")
        run_build("doesnt_exist.yml", "checks_simple.yml", True)
        run_build("build_simple_checks.yml", "doesnt_exist.yml", True)
        run_build("build_simple_checks.yml", None)
        # bad yaml files
        with assertRaisesRegex(self, yaml.parser.ParserError, 'expected'):
            run_build("build_simple_checks.yml", "test_checks.py")
        with assertRaisesRegex(self, yaml.parser.ParserError, 'expected'):
            run_build("test_checks.py", None)

